"""
Test script for file_catalog
"""

from __future__ import absolute_import, division, print_function

from tests.util import unittest_reporter, glob_tests

import logging
logger = logging.getLogger('file_catalog')


try:
    pass
except:
    pass

import unittest
try:
    from unittest.mock import patch
except ImportError:
    pass
import requests_mock
import requests

import iceprod.core.file_catalog
from iceprod.core.jsonUtil import json_encode,json_decode


class file_catalog_test(unittest.TestCase):
    def setUp(self):
        super(file_catalog_test,self).setUp()

    def tearDown(self):
        super(file_catalog_test,self).tearDown()

    @unittest_reporter
    def test_001_FileCatalogLowLevel_init(self):
        url = 'http://foo.bar'
        fc = iceprod.core.file_catalog.FileCatalogLowLevel(url)
        self.assertIn(url, fc.url)

        t = 12.32
        fc = iceprod.core.file_catalog.FileCatalogLowLevel(url, timeout=t)
        self.assertEqual(fc.timeout, t)

    @requests_mock.mock()
    @unittest_reporter
    def test_010_FileCatalogLowLevel_setitem(self, http_mock):
        url = 'http://foo.bar'
        data = url+'/api/file/12345'
        http_mock.get(requests_mock.ANY, exc=KeyError)
        http_mock.post(requests_mock.ANY, content=data.encode('utf-8'))
        fc = iceprod.core.file_catalog.FileCatalogLowLevel(url)
        meta = {'checksum':'thecheck', 'locations':['/path/to/file']}
        fc['blah'] = meta
        self.assertTrue(http_mock.called)
        req = http_mock.request_history[0]
        self.assertEqual(req.method, 'GET')
        req = http_mock.request_history[1]
        self.assertEqual(req.method, 'POST')
        meta_ret = json_decode(req.body)
        self.assertEqual(meta_ret.pop('uid'), 'blah')
        self.assertEqual(meta_ret, meta)

    @requests_mock.mock()
    @unittest_reporter(name='FileCatalogLowLevel_setitem() - update')
    def test_011_FileCatalogLowLevel_setitem(self, http_mock):
        url = 'http://foo.bar'
        data = json_encode({'files':['/api/file/12345']})
        http_mock.get('/api/files', content=data.encode('utf-8'))
        data = url+'/api/file/12345'
        http_mock.put(requests_mock.ANY, content=data.encode('utf-8'))

        fc = iceprod.core.file_catalog.FileCatalogLowLevel(url)
        meta = {'checksum':'thecheck', 'locations':['/path/to/file']}
        fc['blah'] = meta
        self.assertTrue(http_mock.called)
        req = http_mock.request_history[0]
        self.assertEqual(req.method, 'GET')
        req = http_mock.request_history[1]
        self.assertEqual(req.method, 'PUT')
        meta_ret = json_decode(req.body)
        self.assertEqual(meta_ret.pop('uid'), 'blah')
        self.assertEqual(meta_ret, meta)

    @requests_mock.mock()
    @unittest_reporter(name='FileCatalogLowLevel_setitem() - error')
    def test_012_FileCatalogLowLevel_setitem_error(self, http_mock):
        url = 'http://foo.bar'
        http_mock.get(requests_mock.ANY, exc=KeyError)
        http_mock.post(requests_mock.ANY, exc=requests.exceptions.HTTPError)
        fc = iceprod.core.file_catalog.FileCatalogLowLevel(url)
        meta = {'checksum':'thecheck', 'locations':['/path/to/file']}
        try:
            fc['blah'] = meta
        except:
            pass
        else:
            raise Exception('should have raised Exception for non-existance')

        http_mock.post(requests_mock.ANY, exc=requests.exceptions.Timeout)
        try:
            fc['blah'] = meta
        except:
            pass
        else:
            raise Exception('should have raised Exception')

    @requests_mock.mock()
    @unittest_reporter
    def test_020_FileCatalogLowLevel_getitem(self, http_mock):
        data = json_encode({'files':['/api/file/12345']})
        http_mock.get('/api/files', content=data.encode('utf-8'))
        meta = {'uid':'foo','checksum':'thecheck', 'locations':['/path/to/file']}
        data = json_encode(meta)
        http_mock.get('/api/file/12345', content=data.encode('utf-8'))
        url = 'http://foo.bar'
        fc = iceprod.core.file_catalog.FileCatalogLowLevel(url)
        meta_ret = fc['foo']
        self.assertEqual(meta_ret, meta)

    @requests_mock.mock()
    @unittest_reporter(name='FileCatalogLowLevel_getitem() - error')
    def test_021_FileCatalogLowLevel_getitem_error(self, http_mock):
        data = json_encode({'files':[]})
        http_mock.get('/api/files', content=data.encode('utf-8'))
        url = 'http://foo.bar'
        fc = iceprod.core.file_catalog.FileCatalogLowLevel(url)
        try:
            meta = fc['blah']
        except KeyError:
            pass
        else:
            raise Exception('should have raised Exception')

        data = json_encode({'files':['/api/file/12345']})
        http_mock.get('/api/files', content=data.encode('utf-8'))
        http_mock.get('/api/file/12345', exc=requests.exceptions.Timeout)
        try:
            meta = fc['blah']
        except:
            pass
        else:
            raise Exception('should have raised Exception')

        http_mock.get('/api/files', exc=requests.exceptions.HTTPError)
        try:
            meta = fc['blah']
        except:
            pass
        else:
            raise Exception('should have raised Exception')

        http_mock.get('/api/files', exc=requests.exceptions.Timeout)
        try:
            meta = fc['blah']
        except:
            pass
        else:
            raise Exception('should have raised Exception')

    @requests_mock.mock()
    @unittest_reporter
    def test_030_FileCatalogLowLevel_delitem(self, http_mock):
        data = json_encode({'files':['/api/file/12345']})
        http_mock.get('/api/files', content=data.encode('utf-8'))
        http_mock.delete(requests_mock.ANY, content=b'')
        url = 'http://foo.bar'
        fc = iceprod.core.file_catalog.FileCatalogLowLevel(url)
        del fc['foo']
        self.assertTrue(http_mock.called)
        req = http_mock.request_history[0]
        self.assertEqual(req.method, 'GET')
        req = http_mock.request_history[1]
        self.assertEqual(req.method, 'DELETE')

    @requests_mock.mock()
    @unittest_reporter(name='FileCatalogLowLevel_delitem() - error')
    def test_031_FileCatalogLowLevel_delitem_error(self, http_mock):
        data = json_encode({'files':['/api/file/12345']})
        http_mock.get(requests_mock.ANY, content=data.encode('utf-8'))
        http_mock.delete(requests_mock.ANY, exc=requests.exceptions.HTTPError)
        url = 'http://foo.bar'
        fc = iceprod.core.file_catalog.FileCatalogLowLevel(url)
        try:
            del fc['blah']
        except:
            pass
        else:
            raise Exception('should have raised Exception')

        http_mock.get(requests_mock.ANY, content=data.encode('utf-8'))
        http_mock.delete(requests_mock.ANY, exc=requests.exceptions.Timeout)
        try:
            del fc['blah']
        except:
            pass
        else:
            raise Exception('should have raised Exception')

        http_mock.get(requests_mock.ANY, exc=KeyError)
        try:
            del fc['blah']
        except KeyError:
            pass
        else:
            raise Exception('should have raised KeyError')

    @unittest_reporter
    def test_100_FileCatalog_init(self):
        url = 'http://foo.bar'
        fc = iceprod.core.file_catalog.FileCatalog(url)

    @requests_mock.mock()
    @unittest_reporter
    def test_110_FileCatalog_add(self, http_mock):
        url = 'http://foo.bar'
        http_mock.get(requests_mock.ANY, exc=KeyError)
        http_mock.post(requests_mock.ANY, content=b'')
        fc = iceprod.core.file_catalog.FileCatalog(url)

        name = 'foo'
        path = '/path/to/file'
        checksum = 'thechecksum'
        fc.add(name, path, checksum)

        self.assertTrue(http_mock.called)
        req = http_mock.request_history[0]
        self.assertEqual(req.method, 'GET')
        req = http_mock.request_history[1]
        self.assertEqual(req.method, 'POST')
        meta_ret = json_decode(req.body)
        expected = {'uid':name,'locations':[path],'checksum':checksum}
        self.assertEqual(meta_ret, expected)

        metadata = {'foo':'bar'}
        fc.add(name, path, checksum, metadata=metadata)

        self.assertTrue(http_mock.called)
        req = http_mock.request_history[2]
        self.assertEqual(req.method, 'GET')
        req = http_mock.request_history[3]
        self.assertEqual(req.method, 'POST')
        meta_ret = json_decode(req.body)
        expected = {'uid':name,'locations':[path],'checksum':checksum,
                    'foo':'bar'}
        self.assertEqual(meta_ret, expected)

    @requests_mock.mock()
    @unittest_reporter
    def test_120_FileCatalog_get(self, http_mock):
        data = json_encode({'files':['/api/file/12345']})
        http_mock.get('/api/files', content=data.encode('utf-8'))
        url = 'http://foo.bar'
        meta = {'mongo_id':'bar','uid':'foo','checksum':'thecheck', 'locations':['/path/to/file']}
        data = json_encode(meta)
        http_mock.get('/api/file/12345', content=data.encode('utf-8'))
        fc = iceprod.core.file_catalog.FileCatalog(url)

        name = 'foo'
        path, checksum = fc.get(name)

        self.assertTrue(http_mock.called)
        req = http_mock.request_history[0]
        self.assertEqual(req.method, 'GET')
        req = http_mock.request_history[1]
        self.assertEqual(req.method, 'GET')
        self.assertEqual(path, meta['locations'][0])
        self.assertEqual(checksum, meta['checksum'])

    @requests_mock.mock()
    @unittest_reporter
    def test_121_FileCatalog_get_metadata(self, http_mock):
        data = json_encode({'files':['/api/file/12345']})
        http_mock.get('/api/files', content=data.encode('utf-8'))
        url = 'http://foo.bar'
        meta = {'mongo_id':'bar','uid':'foo','checksum':'thecheck', 'locations':['/path/to/file']}
        data = json_encode(meta)
        http_mock.get('/api/file/12345', content=data.encode('utf-8'))
        fc = iceprod.core.file_catalog.FileCatalog(url)

        name = 'foo'
        meta_ret = fc.get_metadata(name)

        self.assertTrue(http_mock.called)
        req = http_mock.request_history[0]
        self.assertEqual(req.method, 'GET')
        req = http_mock.request_history[1]
        self.assertEqual(req.method, 'GET')
        self.assertEqual(meta, meta_ret)

    @requests_mock.mock()
    @unittest_reporter
    def test_130_FileCatalog_delete(self, http_mock):
        url = 'http://foo.bar'
        data = json_encode({'files':['/api/file/12345']})
        http_mock.get('/api/files', content=data.encode('utf-8'))
        http_mock.delete('/api/file/12345', content=b'')
        fc = iceprod.core.file_catalog.FileCatalog(url)

        fc.delete('foo')

        self.assertTrue(http_mock.called)
        req = http_mock.request_history[0]
        self.assertEqual(req.method, 'GET')
        req = http_mock.request_history[1]
        self.assertEqual(req.method, 'DELETE')


def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    alltests = glob_tests(loader.getTestCaseNames(file_catalog_test))
    suite.addTests(loader.loadTestsFromNames(alltests,file_catalog_test))
    return suite



"""
Server for queue management
"""
from collections import defaultdict
from functools import partial
import importlib
import logging
from pathlib import Path
import pkgutil

from prometheus_client import Info, start_http_server
from rest_tools.server import RestServer
from tornado.web import RequestHandler, HTTPError

from iceprod.common.mongo import Mongo
from iceprod.common.prom_utils import AsyncMonitor
from iceprod.s3 import boto3, S3
from iceprod.util import VERSION_STRING
from .config import get_config
from .base_handler import IceProdRestConfig

logger = logging.getLogger('rest-server')


class Error(RequestHandler):
    def prepare(self):
        raise HTTPError(404, 'invalid api route')


class Health(RequestHandler):
    def get(self):
        self.write({})


class Server:
    def __init__(self, s3_override=None):
        config = get_config()

        rest_config = {
            'debug': config.DEBUG,
            'route_stats': {
                'window_size': config.ROUTE_STATS_WINDOW_SIZE,
                'window_time': config.ROUTE_STATS_WINDOW_TIME,
                'timeout': config.ROUTE_STATS_TIMEOUT,
            }
        }
        if config.OPENID_URL:
            logging.info(f'enabling auth via {config.OPENID_URL} for aud "{config.OPENID_AUDIENCE}"')
            rest_config.update({
                'auth': {
                    'openid_url': config.OPENID_URL,
                    'audience': config.OPENID_AUDIENCE,
                }
            })
        elif config.CI_TESTING:
            rest_config.update({
                'auth': {
                    'secret': 'secret',
                }
            })
        else:
            raise RuntimeError('OPENID_URL not specified, and CI_TESTING not enabled!')

        # enable monitoring
        self.prometheus_port = config.PROMETHEUS_PORT if config.PROMETHEUS_PORT > 0 else None
        self.async_monitor = None

        s3conn = None
        if s3_override:
            logging.warning('S3 in testing mode')
            s3conn = S3('', '', '', mock_s3=s3_override)
        elif boto3 and config.S3_ACCESS_KEY and config.S3_SECRET_KEY:
            s3conn = S3(config.S3_ADDRESS, config.S3_ACCESS_KEY, config.S3_SECRET_KEY)
        else:
            logger.warning('S3 is not available!')

        self.db_client = Mongo(
            url=config.DB_URL,
            timeout=config.DB_TIMEOUT,
            write_concern=config.DB_WRITE_CONCERN
        )
        self.indexes = defaultdict(partial(defaultdict, dict))

        kwargs = IceProdRestConfig(rest_config, database=self.db_client.client, s3conn=s3conn)

        server = RestServer(debug=config.DEBUG, max_body_size=config.MAX_BODY_SIZE)

        handler_path = str(Path(__file__).parent / 'handlers')
        for _, name, _ in pkgutil.iter_modules([handler_path]):
            mod = importlib.import_module(f'iceprod.rest.handlers.{name}')
            ret = mod.setup(kwargs)
            for route,cls,kw in ret['routes']:
                kw2 = kw.copy()
                kw2['database'] = kw['database'][ret['database']]
                kw2['db_client'] = self.db_client.client
                server.add_route(route, cls, kw2)
            ii = self.indexes[ret['database']]
            for col in ret['indexes']:
                ii[col].update(ret['indexes'][col])

        server.add_route('/healthz', Health)
        server.add_route(r'/(.*)', Error)

        server.startup(address=config.HOST, port=config.PORT)

        self.server = server

    async def start(self):
        if self.prometheus_port:
            logging.info("starting prometheus on %r", self.prometheus_port)
            start_http_server(self.prometheus_port)
            i = Info('iceprod', 'IceProd information')
            i.info({
                'version': VERSION_STRING,
                'type': 'api',
            })
            self.async_monitor = AsyncMonitor(labels={'type': 'api'})
            await self.async_monitor.start()

        await self.db_client.ping()
        for database, indexes in self.indexes.items():
            await self.db_client.create_indexes(db_name=database, indexes=indexes)

    async def stop(self):
        await self.server.stop()
        if self.async_monitor:
            await self.async_monitor.stop()
        await self.db_client.close()



import asyncio
import logging

import pytest
import requests.exceptions
from rest_tools.utils.json_util import json_decode


async def test_rest_task_stats_post(server):
    client = server(roles=['system'])

    task_id = 'bar'
    data = {
        'dataset_id': 'foo',
        'bar': 1.23456,
        'baz': [1,2,3,4],
    }
    ret = await client.request('POST', f'/tasks/{task_id}/task_stats', data)
    task_stat_id = ret['result']


async def test_rest_task_stats_post_bad_role(server):
    client = server(roles=['user'])

    task_id = 'bar'
    data = {
        'dataset_id': 'foo',
        'bar': 1.23456,
        'baz': [1,2,3,4],
    }
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', f'/tasks/{task_id}/task_stats', data)
    assert exc_info.value.response.status_code == 403


async def test_rest_task_stats_bulk(server):
    client = server(roles=['system'])

    task_id = 'bar'
    data = {
        'dataset_id': 'foo',
        'bar': 1.23456,
        'baz': [1,2,3,4],
    }
    ret = await client.request('POST', f'/tasks/{task_id}/task_stats', data)
    task_stat_id = ret['result']
    ret = await client.request('POST', f'/tasks/{task_id}/task_stats', data)
    task_stat_id2 = ret['result']
    ret = await client.request('POST', f'/tasks/{task_id}/task_stats', data)
    task_stat_id3 = ret['result']

    url, kwargs = client._prepare('GET', f'/datasets/{data["dataset_id"]}/bulk/task_stats', {'buffer_size': 2})
    ret = await asyncio.wrap_future(client.session.request('GET', url, **kwargs))
    ret.raise_for_status()
    logging.info('ret.content: %r', ret.content)
    task_stats = [json_decode(r) for r in ret.content.split(b'\n') if r.strip()]
    ret_task_ids = [t['task_stat_id'] for t in task_stats]
    assert ret_task_ids == [task_stat_id, task_stat_id2, task_stat_id3]


async def test_rest_task_stats_get(server):
    client = server(roles=['system'])

    task_id = 'bar'
    data = {
        'dataset_id': 'foo',
        'bar': 1.23456,
        'baz': [1,2,3,4],
    }
    data_stat = data.copy()
    del data_stat['dataset_id']
    ret = await client.request('POST', f'/tasks/{task_id}/task_stats', data)
    task_stat_id = ret['result']
    ret = await client.request('POST', f'/tasks/{task_id}/task_stats', data)
    task_stat_id2 = ret['result']

    ret = await client.request('GET', f'/datasets/{data["dataset_id"]}/tasks/{task_id}/task_stats')
    assert len(ret) == 2
    assert task_stat_id in ret
    assert task_stat_id2 in ret
    assert 'task_id' in ret[task_stat_id]
    assert task_id == ret[task_stat_id]['task_id']
    assert data_stat == ret[task_stat_id]['stats']

    ret = await client.request('GET', f'/datasets/{data["dataset_id"]}/tasks/{task_id}/task_stats', {'last': 'true'})
    assert len(ret) == 1
    assert task_stat_id2 in ret
    assert 'task_id' in ret[task_stat_id2]
    assert task_id == ret[task_stat_id2]['task_id']
    assert data_stat == ret[task_stat_id2]['stats']

    ret = await client.request('GET', f'/datasets/{data["dataset_id"]}/tasks/{task_id}/task_stats', {'last': 'true', 'keys': 'task_id'})
    assert len(ret) == 1
    assert task_stat_id2 in ret
    assert 'task_id' in ret[task_stat_id2]
    assert task_id == ret[task_stat_id2]['task_id']
    assert 'stats' not in ret[task_stat_id2]


async def test_rest_task_stats_get_details(server):
    client = server(roles=['system'])

    task_id = 'bar'
    data = {
        'dataset_id': 'foo',
        'bar': 1.23456,
        'baz': [1,2,3,4],
    }
    data_stat = data.copy()
    del data_stat['dataset_id']
    ret = await client.request('POST', f'/tasks/{task_id}/task_stats', data)
    task_stat_id = ret['result']

    ret = await client.request('GET', f'/datasets/{data["dataset_id"]}/tasks/{task_id}/task_stats/{task_stat_id}')
    assert task_stat_id == ret['task_stat_id']
    assert task_id == ret['task_id']
    assert data_stat == ret['stats']



import asyncio
import logging
import time

import httpx
import jwt

from iceprod.common.mongo import AsyncDatabase
from .util import ClientCreds, get_expiration

logger = logging.getLogger('refresh_service')


class ExchangeException(Exception):
    pass


class RefreshService:
    """
    OAuth refresh service

    Args:
        database: mongo database
        clients: json string of {url: [client id, client secret]}
        refresh_window: how long after last use to keep refreshing (in hours)
        expire_buffer: how long before expiration to refresh (in minutes)
        service_run_interval: seconds between refresh runs
    """
    def __init__(self, database: AsyncDatabase, clients: str, refresh_window: float, expire_buffer: int, service_run_interval: float):
        self.db = database
        self.clients = ClientCreds(clients)
        self.clients.validate()
        self.refresh_window = refresh_window * 3600.
        self.expire_buffer = expire_buffer

        self.start_time = time.time()
        self.service_run_interval = service_run_interval
        self.last_run_time = None
        self.last_success_time = None

    async def refresh_cred(self, cred):
        if not cred.get('refresh_token'):
            raise Exception('cred does not have a refresh token')

        openid_url = jwt.decode(cred['refresh_token'], options={"verify_signature": False})['iss']
        try:
            client = self.clients.get_client(openid_url)
        except KeyError:
            raise Exception('jwt issuer not registered')

        # try the refresh token
        args = {
            'grant_type': 'refresh_token',
            'refresh_token': cred['refresh_token'],
            'client_id': client.client_id,
        }
        if client.client_secret:
            args['client_secret'] = client.client_secret
        if cred.get('scope', None) is not None:
            args['scope'] = cred['scope']

        logging.warning('refreshing on %s with args %r', client.auth.token_url, args)

        new_cred = {}
        try:
            async with httpx.AsyncClient() as http_client:
                r = await http_client.post(client.auth.token_url, data=args)
            r.raise_for_status()
            req = r.json()
        except httpx.HTTPStatusError as exc:
            logger.debug('%r', exc.response.text)
            try:
                req = exc.response.json()
            except Exception:
                req = {}
            error = req.get('error', '')
            raise Exception(f'Refresh request failed: {error}') from exc
        else:
            logger.debug('OpenID token refreshed')
            new_cred['access_token'] = req['access_token']
            new_cred['refresh_token'] = req['refresh_token']
            new_cred['expiration'] = get_expiration(req['access_token'])
            new_cred['scope'] = req.get('scope', cred.get('scope', ''))
            logger.debug('%r', new_cred)

        return new_cred

    async def exchange_cred(self, cred, *, client_id: str, new_scope: str | None = None):
        """Exchange a refresh token for one on another client, and return that."""
        if not cred.get('refresh_token'):
            raise Exception('cred does not have a refresh token')
        if not cred.get('access_token'):
            raise Exception('cred does not have an access token')

        openid_url = jwt.decode(cred['refresh_token'], options={"verify_signature": False})['iss']
        try:
            client = self.clients.get_client(openid_url)
        except KeyError:
            raise Exception('jwt issuer not registered')

        # try the refresh token
        args = {
            'grant_type': 'urn:ietf:params:oauth:grant-type:token-exchange',
            'client_id': client.client_id,
            'audience': client_id,  # target client
            'scope': new_scope if new_scope else cred['scope'],
            'subject_token': cred['access_token'],
            'subject_token_type': 'urn:ietf:params:oauth:token-type:access_token',
        }
        if client.client_secret:
            args['client_secret'] = client.client_secret

        logging.warning('exchanging on %s with args %r', client.auth.token_url, args)

        new_cred = cred.copy()
        try:
            async with httpx.AsyncClient() as http_client:
                r = await http_client.post(client.auth.token_url, data=args)
            r.raise_for_status()
            req = r.json()
        except httpx.HTTPStatusError as exc:
            logger.debug('%r', exc.response.text)
            try:
                req = exc.response.json()
            except Exception:
                req = {}
            error = req.get('error', '')
            desc = req.get('error_description', '')
            raise ExchangeException(f'Exchange request failed: {error} - {desc}') from exc
        else:
            logger.debug('OpenID token exchanged')
            new_cred['access_token'] = req['access_token']
            new_cred['refresh_token'] = req['refresh_token']
            new_cred['expiration'] = get_expiration(req['access_token'])
            new_cred['scope'] = req.get('scope', args.get('scope', ''))
            logger.debug('%r', new_cred)

        return new_cred

    async def create_cred(self, *, url: str, transfer_prefix: str, username: str, scope: str):
        """Do an impersonation token exchange workflow to generate a cred for a user."""
        try:
            client = self.clients.get_client(url)
        except KeyError:
            raise Exception('url not registered')
        if transfer_prefix not in client.transfer_prefix:
            raise Exception('client transfer prefix does not match')

        # try the refresh token
        args = {
            'grant_type': 'urn:ietf:params:oauth:grant-type:token-exchange',
            'client_id': client.client_id,
            'scope': scope,
            'requested_subject': username
        }
        if client.client_secret:
            args['client_secret'] = client.client_secret

        logging.warning('exchanging on %s with args %r', client.auth.token_url, args)

        new_cred = {
            'url': url,
            'type': 'oauth',
            'transfer_prefix': transfer_prefix,
            'scope': scope,
        }
        try:
            async with httpx.AsyncClient() as http_client:
                r = await http_client.post(client.auth.token_url, data=args)
            r.raise_for_status()
            req = r.json()
        except httpx.HTTPStatusError as exc:
            logger.debug('%r', exc.response.text)
            try:
                req = exc.response.json()
            except Exception:
                req = {}
            error = req.get('error', '')
            desc = req.get('error_description', '')
            raise Exception(f'Exchange request failed: {error} - {desc}') from exc
        else:
            logger.debug('OpenID token exchanged')
            new_cred['access_token'] = req['access_token']
            new_cred['refresh_token'] = req['refresh_token']
            new_cred['expiration'] = get_expiration(req['access_token'])
            if 'scope' in req and req['scope'] != scope:
                raise Exception('scopes do not match!')
            logger.debug('%r', new_cred)

        return new_cred

    def should_refresh(self, cred):
        logger.info('should_refresh for cred %r', cred)
        now = time.time()
        # refresh_exp = now + self.expire_buffer
        last_use_date = now - self.refresh_window

        if cred.get('last_use', 0) < last_use_date:
            return False
        return True

    async def run(self):
        """
        Run loop.
        """
        while True:
            await self._run_once()
            await asyncio.sleep(self.service_run_interval)

    async def _run_once(self):
        try:
            self.last_run_time = time.time()
            last_use_check = self.last_run_time - self.refresh_window
            exp_check = self.last_run_time + self.expire_buffer
            filters = {
                'type': 'oauth',
                'last_use': {'$gt': last_use_check},
                'expiration': {'$lt': exp_check},
                'scope': {'$exists': True},
            }

            await self._run_once_user(filters)
            await self._run_once_group(filters)
            await self._run_once_dataset(filters)

            self.last_success_time = time.time()
        except Exception:
            logger.error('error running refresh', exc_info=True)

    async def _run_once_user(self, filters):
        user_creds = {}
        async for row in self.db.user_creds.find(filters, {'_id': False}):
            user_creds[row['url']] = row

        for cred in user_creds.values():
            if cred['type'] != 'oauth':
                continue
            if not cred['refresh_token']:
                logger.info('skipping non-refresh token for user %s, url %s', cred['username'], cred['url'])
                continue
            try:
                if self.should_refresh(cred):
                    args = await self.refresh_cred(cred)
                    await self.db.user_creds.update_one({'username': cred['username'], 'url': cred['url'], 'scope': cred['scope']}, {'$set': args})
                    logger.info('refreshed token for user %s, url %s, scope %s', cred['username'], cred['url'], cred['scope'])
                else:
                    logger.info('not yet time to refresh token for user %s, url %s, scope %s', cred['username'], cred['url'], cred['scope'])
            except Exception:
                logger.error('error refreshing token for user %s, url: %s, scope %s', cred['username'], cred['url'], cred['scope'], exc_info=True)

    async def _run_once_group(self, filters):
        group_creds = {}
        async for row in self.db.group_creds.find(filters, {'_id': False}):
            group_creds[row['url']] = row

        for cred in group_creds.values():
            if cred['type'] != 'oauth':
                continue
            if not cred['refresh_token']:
                logger.info('skipping non-refresh token for group %s, url %s', cred['groupname'], cred['url'])
                continue
            try:
                if self.should_refresh(cred):
                    args = await self.refresh_cred(cred)
                    await self.db.group_creds.update_one({'groupname': cred['groupname'], 'url': cred['url']}, {'$set': args})
                    logger.info('refreshed token for group %s, url %s', cred['groupname'], cred['url'])
                else:
                    logger.info('not yet time to refresh token for group %s, url %s', cred['groupname'], cred['url'])
            except Exception:
                logger.error('error refreshing token for group %s, url: %s', cred['groupname'], cred['url'], exc_info=True)

    async def _run_once_dataset(self, filters):
        dataset_creds = []
        async for row in self.db.dataset_creds.find(filters, {'_id': False}):
            dataset_creds.append(row)

        for cred in dataset_creds:
            if cred['type'] != 'oauth':
                continue
            if not cred['refresh_token']:
                logger.info('skipping non-refresh token for dataset %s, task, %s, url %s', cred['dataset_id'], cred.get('task_name',''), cred['url'])
                continue
            try:
                if self.should_refresh(cred):
                    args = await self.refresh_cred(cred)
                    cred_filter = {'dataset_id': cred['dataset_id'], 'scope': cred['scope'], 'url': cred['url']}
                    if task_name := cred.get('task_name', ''):
                        cred_filter['task_name'] = task_name
                    await self.db.dataset_creds.update_one(cred_filter, {'$set': args})
                    logger.info('refreshed token for dataset %s, task, %s, url %s', cred['dataset_id'], cred.get('task_name',''), cred['url'])
                else:
                    logger.info('not yet time to refresh token for dataset %s, task, %s, url %s', cred['dataset_id'], cred.get('task_name',''), cred['url'])
            except Exception:
                logger.error('error refreshing token for dataset %s, task, %s, url %s', cred['dataset_id'], cred.get('task_name',''), cred['url'], exc_info=True)



import json
import time
from unittest.mock import MagicMock

import jwt
import pytest

import iceprod.credentials.util


def test_credentials_util_get_expiration():
    t = time.time()
    tok = jwt.encode({'exp': t}, 'secret')
    e = iceprod.credentials.util.get_expiration(tok)
    assert t == e


def test_credentials_util_is_expired():
    cred = {
        'type': 'oauth',
        'expiration': time.time()
    }
    assert iceprod.credentials.util.is_expired(cred)

    cred = {
        'type': 'oauth',
        'expiration': time.time()+100
    }
    assert not iceprod.credentials.util.is_expired(cred)


def test_client_validate(monkeypatch):
    authmock = MagicMock()
    authmock.return_value.token_url = 'http://iceprod.test'
    monkeypatch.setattr(iceprod.credentials.util, '_get_auth', authmock)

    invalid = json.dumps({
        'http://iceprod.test': []
    })

    with pytest.raises(Exception):
        iceprod.credentials.util.ClientCreds(invalid).validate()

    valid =  json.dumps({
        'http://iceprod.test': {'client_id': 'id', 'client_secret': 'secret'}
    })
    cc = iceprod.credentials.util.ClientCreds(valid)
    cc.validate()
    assert cc.get_client('http://iceprod.test').client_id == 'id'
    assert cc.get_client('http://iceprod.test').client_secret == 'secret'

    valid =  json.dumps({
        'http://iceprod.test': {'client_id': 'id'}
    })
    cc = iceprod.credentials.util.ClientCreds(valid)
    cc.validate()
    assert cc.get_client('http://iceprod.test').client_id == 'id'
    assert cc.get_client('http://iceprod.test').client_secret == ''



import logging
from typing import Any

from rest_tools.server import RestHandlerSetup, RestHandler
from tornado.escape import json_encode

from ..common.mongo import AsyncMongoClient, AsyncDatabase
from iceprod.util import VERSION_STRING
from iceprod.common.prom_utils import PromRequestMixin
from .auth import AttrAuthMixin

logger = logging.getLogger('rest')

type DB = AsyncMongoClient | AsyncDatabase


def IceProdRestConfig(config: dict[str, Any], database: DB | None = None, auth_database: AsyncDatabase | None = None, s3conn=None):
    if config:
        config['server_header'] = 'IceProd/' + VERSION_STRING
    ret = RestHandlerSetup(config)
    ret['database'] = database
    ret['s3'] = s3conn
    return ret


class APIBase(AttrAuthMixin, PromRequestMixin, RestHandler):
    """Default REST handler"""
    def initialize(self, *args, database: DB, db_client: AsyncMongoClient | None = None, s3=None, **kwargs):  # type: ignore[override]
        logger.info('initialze APIBase: args=%r, kwargs=%r', args, kwargs)
        super().initialize(*args, **kwargs)
        logger.info('do rest of initialize APIBase')
        self.db = database
        self.db_client = db_client
        self.auth_db: AsyncDatabase | None = db_client['auth'] if db_client else None  # type: ignore
        self.s3 = s3

    def get_template_namespace(self):
        namespace = super().get_template_namespace()
        namespace['version'] = VERSION_STRING
        return namespace

    def get_current_user(self):
        """Get keycloak username if available"""
        if not super().get_current_user():
            return None

        if 'iceprod-system' in self.auth_data.get('resource_access', {}).get('iceprod', {}).get('roles', []):
            username = 'iceprod-system'
        else:
            username = self.auth_data.get('preferred_username', None)
            if not username:
                logger.info('could not find auth username')

        return username

    def prepare(self):
        super().prepare()
        # Set the Cache-Control header to "no-store"
        self.set_header('Cache-Control', 'no-store, no-cache, must-revalidate, max-age=0')
        # Also set Pragma and Expires headers for compatibility with older HTTP versions
        self.set_header('Pragma', 'no-cache')
        self.set_header('Expires', '0')

    def write(self, chunk: str | bytes | dict | list) -> None:  # type: ignore[override]
        """Write dict or list to json"""
        if isinstance(chunk, (dict, list)):
            self.set_header("Content-Type", "application/json; charset=UTF-8")
            super().write(json_encode(chunk))
        else:
            super().write(chunk)



#!/usr/bin/env python3

"""
This is a custom HTCondor file transfer plugin for IceProd.
In this example, it transfers files described by a
iceprod-plugin://transfer-proto://path/to/file?mapping=filename metadata-URL,
by copying them from the path indicated to a job's working directory.
"""

import glob
import os
import sys
import io
import subprocess
import time
import traceback

try:
    import requests
except ImportError:
    requests = None  # type: ignore

try:
    from classad import ClassAd, parseAds  # type: ignore
except ImportError:
    import re
    import json

    class ClassAd(dict):  # type: ignore
        def printOld(self):
            ret = []
            for k,v in self.items():
                if isinstance(v, str):
                    v = '"{0}"'.format(v)
                ret.append('{0} = {1}'.format(k, v))
            return '[\n' + ';\n'.join(ret) + '\n]\n'

    def parseAds(data):
        if isinstance(data, io.IOBase):
            data = data.read()
        ret = []
        for ad in re.findall(r'\s*?\[([\w\W]*?)\]', data):
            ad_ret = {}
            for k,v in re.findall(r'\s*?(\w+) += +([^;\n]+);?', ad):
                ad_ret[k] = json.loads(v)
            ret.append(ad_ret)
        return ret


DEFAULT_TIMEOUT = 300
PLUGIN_VERSION = '1.0.0'

EXIT_SUCCESS = 0
EXIT_FAILURE = 1
EXIT_AUTHENTICATION_REFRESH = 2
EXIT_ARGPARSE = 3
EXIT_INFILES = 4
EXIT_APP = 10


class GridFTPError(RuntimeError):
    pass


def print_help(stream=sys.stderr):
    help_msg = '''Usage: {0} -infile <input-filename> -outfile <output-filename>
       {0} -classad

Options:
  -classad                    Print a ClassAd containing the capablities of this
                              file transfer plugin.
  -infile <input-filename>    Input ClassAd file
  -outfile <output-filename>  Output ClassAd file
  -upload                     Indicates this transfer is an upload (default is
                              download)
'''
    stream.write(help_msg.format(sys.argv[0]))


def print_capabilities():
    capabilities = {
        'MultipleFileSupport': True,
        'PluginType': 'FileTransfer',
        # SupportedMethods indicates which URL methods/types this plugin supports
        'SupportedMethods': 'iceprod-plugin',
        'Version': PLUGIN_VERSION,
    }
    sys.stdout.write(ClassAd(capabilities).printOld())


def parse_args():

    # The only argument lists that are acceptable are
    # <this> -classad
    # <this> -infile <input-filename> -outfile <output-filename>
    # <this> -outfile <output-filename> -infile <input-filename>
    if len(sys.argv) not in [2, 5, 6]:
        print_help()
        sys.exit(EXIT_FAILURE)

    # If -classad, print the capabilities of the plugin and exit early
    if (len(sys.argv) == 2) and (sys.argv[1] == '-classad'):
        print_capabilities()
        sys.exit(EXIT_SUCCESS)

    # If -upload, set is_upload to True and remove it from the args list
    is_upload = False
    if '-upload' in sys.argv[1:]:
        is_upload = True
        sys.argv.remove('-upload')

    # -infile and -outfile must be in the first and third position
    if not (
            ('-infile' in sys.argv[1:]) and
            ('-outfile' in sys.argv[1:]) and
            (sys.argv[1] in ['-infile', '-outfile']) and
            (sys.argv[3] in ['-infile', '-outfile']) and
            (len(sys.argv) == 5)):
        print_help()
        sys.exit(1)
    infile = None
    outfile = None
    try:
        for i, arg in enumerate(sys.argv):
            if i == 0:
                continue
            elif arg == '-infile':
                infile = sys.argv[i+1]
            elif arg == '-outfile':
                outfile = sys.argv[i+1]
    except IndexError:
        print_help()
        sys.exit(EXIT_FAILURE)

    return {'infile': infile, 'outfile': outfile, 'upload': is_upload}


def format_error(error):
    return '{0}: {1}'.format(type(error).__name__, str(error))


def get_error_dict(error, url=''):
    error_string = format_error(error)
    error_dict = {
        'TransferSuccess': False,
        'TransferError': error_string,
        'TransferUrl': url,
    }

    return error_dict


class IceProdPlugin:

    def setup_env(self):
        # if a proxy exists, use it
        proxies = glob.glob(os.path.join(os.getcwd(), 'x509up_*'))
        if proxies:
            os.environ['X509_USER_PROXY'] = proxies[0]
        else:
            raise RuntimeError('X509_USER_PROXY does not exist')

        if not os.path.exists('/cvmfs/icecube.opensciencegrid.org/iceprod/v2.7.1/env-shell.sh'):
            raise RuntimeError('CVMFS does not exist')

    def _split_url(self, url):
        # strip off iceprod-plugin:// prefix
        url = url.split('://',1)[-1]
        # split url
        method = url.split('://',1)[0]
        transfer = 'true'
        mapping = None
        if '-' in method:
            transfer, method = method.split('-',1)
            url = url.split('-',1)[-1]
        if '?' in url:
            url, args = url.split('?',1)
            args = {x.split('=')[0]: x.split('=',1)[-1] for x in args.split('&')}
            mapping = args.get('mapping')
        return {
            'method': method,
            'url': url,
            'transfer': transfer,
            'mapping': mapping,
        }

    def _do_globus_transfer(self, inpath, outpath):
        apptainer = []
        try:
            apptainer_location = subprocess.check_output('which apptainer', shell=True).decode('utf-8').strip()
            apptainer = [apptainer_location, 'run', '-B/cvmfs', '/cvmfs/singularity.opensciencegrid.org/opensciencegrid/osgvo-el7:latest']
        except Exception:
            apptainer = []
        try:
            subprocess.check_output(apptainer + [
                '/cvmfs/icecube.opensciencegrid.org/iceprod/v2.7.1/env-shell.sh',
                'globus-url-copy',
                '-cd',
                '-rst',
                inpath,
                outpath,
            ], stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            if e.output:
                for line in e.output.decode('utf-8').split('\n'):
                    if line.lower().startswith('error'):
                        raise GridFTPError('globus-url-copy failed: '+line)
            raise

    def download_file(self, url, local_file_path):

        start_time = time.time()
        file_size = 0

        # Download transfer logic goes here
        ret = self._split_url(url)
        url = ret['url']
        method = ret['method']
        transfer = ret['transfer']
        mapping = ret['mapping']

        if transfer in ('true', 'maybe'):
            if method == 'gsiftp':
                try:
                    self._do_globus_transfer(url, 'file://'+os.path.abspath(local_file_path))
                    file_size = os.stat(local_file_path).st_size
                except GridFTPError:
                    if transfer != 'maybe':
                        raise

            elif method in ('http', 'https'):
                response = requests.get(url, stream=True, timeout=DEFAULT_TIMEOUT)
                try:
                    response.raise_for_status()
                    try:
                        content_length = int(response.headers['Content-Length'])
                    except (ValueError, KeyError):
                        content_length = False
                    with open(local_file_path, 'wb') as f:
                        file_size = 0
                        for chunk in response.iter_content(chunk_size=8192):
                            file_size += len(chunk)
                            f.write(chunk)
                    file_size = content_length or file_size
                except Exception as err:
                    if not (transfer == 'maybe' and response.status_code == 404):
                        # skip error if transfer=maybe and the file wasn't found
                        raise err
                finally:
                    response.close()

            else:
                raise Exception('unknown protocol "{0}"'.format(method))

            if mapping and os.path.exists(local_file_path):
                os.rename(os.path.basename(local_file_path), mapping)

        end_time = time.time()

        # Get transfer statistics
        transfer_stats = {
            'TransferSuccess': True,
            'TransferProtocol': method,
            'TransferType': 'upload',
            'TransferFileName': local_file_path,
            'TransferFileBytes': file_size,
            'TransferTotalBytes': file_size,
            'TransferStartTime': int(start_time),
            'TransferEndTime': int(end_time),
            'ConnectionTimeSeconds': end_time - start_time,
            'TransferUrl': url,
        }
        if mapping:
            transfer_stats['MappedFileName'] = mapping

        return transfer_stats

    def upload_file(self, url, local_file_path):

        start_time = time.time()
        file_size = 0

        # Upload transfer logic goes here
        ret = self._split_url(url)
        url = ret['url']
        method = ret['method']
        transfer = ret['transfer']

        if transfer == 'true' or (transfer == 'maybe' and os.path.exists(local_file_path)):
            if method == 'gsiftp':
                self._do_globus_transfer('file://'+os.path.abspath(local_file_path), url)
                file_size = os.stat(local_file_path).st_size

            elif method in ('http', 'https'):
                with open(local_file_path, 'rb') as f:
                    response = requests.put(url, data=f, timeout=DEFAULT_TIMEOUT)
                    response.raise_for_status()
                file_size = os.stat(local_file_path).st_size

            else:
                raise Exception('unknown protocol "{0}"'.format(method))

        end_time = time.time()

        # Get transfer statistics
        transfer_stats = {
            'TransferSuccess': True,
            'TransferProtocol': method,
            'TransferType': 'upload',
            'TransferFileName': local_file_path,
            'TransferFileBytes': file_size,
            'TransferTotalBytes': file_size,
            'TransferStartTime': int(start_time),
            'TransferEndTime': int(end_time),
            'ConnectionTimeSeconds': end_time - start_time,
            'TransferUrl': url,
        }

        return transfer_stats


if __name__ == '__main__':

    # Start by parsing input arguments
    try:
        args = parse_args()
    except Exception:
        with open('_condor_stdout', 'w') as f:
            traceback.print_exc(file=f)
        sys.exit(EXIT_ARGPARSE)

    iceprod_plugin = IceProdPlugin()
    iceprod_plugin.setup_env()

    # Parse in the classads stored in the input file.
    # Each ad represents a single file to be transferred.
    try:
        infile_ads = parseAds(open(args['infile'], 'r'))
    except Exception as err:
        with open('_condor_stdout', 'w') as f:
            traceback.print_exc(file=f)
        try:
            with open(args['outfile'], 'w') as outfile:
                outfile_dict = get_error_dict(err)
                outfile.write(str(ClassAd(outfile_dict)))
        except Exception:
            pass
        sys.exit(EXIT_INFILES)

    # Now iterate over the list of classads and perform the transfers.
    try:
        with open(args['outfile'], 'w') as outfile:
            for ad in infile_ads:
                try:
                    if not args['upload']:
                        outfile_dict = iceprod_plugin.download_file(ad['Url'], ad['LocalFileName'])
                    else:
                        outfile_dict = iceprod_plugin.upload_file(ad['Url'], ad['LocalFileName'])

                    outfile.write(str(ClassAd(outfile_dict)))

                except Exception as err:
                    with open('_condor_stdout', 'w') as f:
                        traceback.print_exc(file=f)
                    try:
                        outfile_dict = get_error_dict(err, url=ad['Url'])
                        outfile.write(str(ClassAd(outfile_dict)))
                    except Exception:
                        pass
                    sys.exit(EXIT_APP)

    except Exception:
        with open('_condor_stdout', 'w') as f:
            traceback.print_exc(file=f)
        sys.exit(EXIT_FAILURE)



"""
Test script for resources
"""

import logging
import os

import iceprod.core.resources


logger = logging.getLogger('resources')

 
def test_000_Resources():
    r = iceprod.core.resources.Resources
    for t in ('cpu','gpu','memory','disk','time'):
        assert t in r.defaults


def test_230_du(tmp_path):
    du_dir = str(tmp_path / 'test')
    os.mkdir(du_dir)
    for f in ('a','b','c'):
        path = os.path.join(du_dir,f)
        open(path,'w').write('a'*100)
    assert iceprod.core.resources.du(du_dir) == 300


def test_231_du_symlink(tmp_path):
    du_dir = str(tmp_path / 'test')
    os.mkdir(du_dir)
    for f in ('a','b','c'):
        path = os.path.join(du_dir,f)
        open(path,'w').write('a'*100)
    os.symlink(os.path.join(du_dir,'a'), os.path.join(du_dir,'l'))
    assert iceprod.core.resources.du(du_dir) == 300


def test_232_du_dir_symlink(tmp_path):
    du_dir = str(tmp_path / 'test')
    os.mkdir(du_dir)
    for f in ('a','b','c'):
        path = os.path.join(du_dir,f)
        open(path,'w').write('a'*100)
    os.symlink(os.path.join(du_dir,'a'), os.path.join(du_dir,'l'))
    os.mkdir(os.path.join(du_dir,'subdir'))
    for f in ('a','b','c'):
        path = os.path.join(du_dir,'subdir',f)
        open(path,'w').write('a'*100)
    os.symlink(os.path.join(du_dir,'subdir'), os.path.join(du_dir,'s2'))
    os.symlink(os.path.join(du_dir,'subdir','a'), os.path.join(du_dir,'subdir','s3'))
    assert iceprod.core.resources.du(du_dir) == 600


def test_300_group_hasher():
    r = iceprod.core.resources.Resources.defaults
    h = iceprod.core.resources.group_hasher(r)


def test_301_group_hasher():
    r = iceprod.core.resources.Resources.defaults.copy()

    hashes = set()
    for i in range(1,100):
        r['memory'] = i
        hashes.add(iceprod.core.resources.group_hasher(r))
    logger.info('hashes: %r', hashes)
    assert len(hashes) < 15


def test_400_sanitized_requirements():
    r = {'cpu':2,'gpu':1}
    ret = iceprod.core.resources.sanitized_requirements(r)
    assert r['cpu'] == ret['cpu']
    assert r['gpu'] == ret['gpu']
    assert 'memory' not in ret

    r = {'os':['RHEL_7_x86_64'], 'site':'foo'}
    ret = iceprod.core.resources.sanitized_requirements(r)
    assert r['os'] == ret['os']
    assert r['site'] == ret['site']
    assert 'gpu' not in ret
    assert 'memory' not in ret


def test_401_sanitized_requirements():
    r = {'cpu':2,'gpu':1}
    ret = iceprod.core.resources.sanitized_requirements(r, use_defaults=True)
    assert r['cpu'] == ret['cpu']
    assert r['gpu'] == ret['gpu']
    assert iceprod.core.resources.Resources.defaults['memory'] == ret['memory']

    r = {'os':['RHEL_7_x86_64'], 'site':'foo'}
    ret = iceprod.core.resources.sanitized_requirements(r, use_defaults=True)
    assert r['os'] == ret['os']
    assert r['site'] == ret['site']
    assert 'gpu' not in ret
    assert iceprod.core.resources.Resources.defaults['memory'] == ret['memory']



"""
Test script for config
"""

import os
import json

import pytest

import iceprod.server.config


def test_01_IceProdConfig():
    """Test config.IceProdConfig()"""
    cfg = iceprod.server.config.IceProdConfig(defaults=False, validate=False)
    cfg.save()
    if not os.path.exists(cfg.filename):
        raise Exception('did not save configfile')

    with open(cfg.filename) as f:
        assert f.read() == '{}'

    cfg['testing'] = [1,2,3]
    assert cfg['testing'] == [1,2,3]

    expected = dict(cfg)
    with open(cfg.filename) as f:
        actual = json.load(f)
    assert actual == expected

    cfg.load()
    assert cfg['testing'] == [1,2,3]

    del cfg['testing']
    assert 'testing' not in cfg

    expected = {}
    with open(cfg.filename) as f:
        actual = json.load(f)
    assert actual == expected


def test_02_filename(i3prod_path):
    """Test config.IceProdConfig()"""
    name = str(i3prod_path / 'test.json')
    cfg = iceprod.server.config.IceProdConfig(filename=name)
    assert cfg.filename == name


def test_10_config_override():
    vals = ['test=foo']
    cfg = iceprod.server.config.IceProdConfig(override=vals, defaults=False, validate=False)
    assert cfg['test'] == 'foo'
    del cfg['test']

    vals = ['test.test2.test3=123', 'test.test4=456.5', 'test2={"foo":123}','test3=true']
    cfg = iceprod.server.config.IceProdConfig(override=vals, defaults=False, validate=False)
    assert cfg['test']['test2']['test3'] == 123
    assert cfg['test']['test4'] == 456.5
    assert cfg['test2'] == {'foo': 123}
    assert cfg['test3'] is True

    vals = ['queue.resources.cpu=1', 'queue.exclusive=true']
    cfg = iceprod.server.config.IceProdConfig(override=vals, defaults=False, validate=False)


def test_20_defaults():
    cfg = iceprod.server.config.IceProdConfig(validate=False)
    assert cfg['logging']['level'] == 'INFO'


def test_30_validate():
    cfg = iceprod.server.config.IceProdConfig()

    cfg['logging']['level'] = 'foo'
    with pytest.raises(Exception):
        cfg.do_validate()


def test_40_save():
    cfg = iceprod.server.config.IceProdConfig(validate=False, save=False)
    assert os.path.exists(cfg.filename) is False





async def test_rest_grids_empty(server):
    client = server(roles=['system'])

    ret = await client.request('GET', '/grids')
    assert ret == {}


async def test_rest_grids_add(server):
    client = server(roles=['system'])

    data = {
        'host': 'foo.bar.baz',
        'queues': {'foo': 'HTCondor'},
        'version': '1.2.3',
    }
    ret = await client.request('POST', '/grids', data)
    grid_id = ret['result']

    ret = await client.request('GET', '/grids')
    assert grid_id in ret
    for k in data:
        assert k in ret[grid_id]
        assert data[k] == ret[grid_id][k]


async def test_rest_grids_details(server):
    client = server(roles=['system'])

    data = {
        'host': 'foo.bar.baz',
        'queues': {'foo': 'HTCondor'},
        'version': '1.2.3',
    }
    ret = await client.request('POST', '/grids', data)
    grid_id = ret['result']

    ret = await client.request('GET', f'/grids/{grid_id}')
    for k in data:
        assert k in ret
        assert data[k] == ret[k]

async def test_rest_grids_patch(server):
    client = server(roles=['system'])

    data = {
        'host': 'foo.bar.baz',
        'queues': {'foo': 'HTCondor'},
        'version': '1.2.3',
    }
    ret = await client.request('POST', '/grids', data)
    grid_id = ret['result']

    new_data = {
        'queues': {'foo': 'HTCondor', 'bar': 'HTCondor'},
        'version': '1.2.8',
    }
    ret = await client.request('PATCH', f'/grids/{grid_id}', new_data)
    
    ret2 = await client.request('GET', f'/grids/{grid_id}')
    assert ret == ret2
    for k in new_data:
        assert k in ret
        assert new_data[k] == ret[k]



"""
Test Server
"""

from __future__ import absolute_import, division, print_function


import logging
logger = logging.getLogger('iceprod_server_test')

import os
import glob

from unittest.mock import AsyncMock, MagicMock

import pytest
import iceprod.server.server
from iceprod.server.server import Server, roll_files


@pytest.fixture(autouse=True)
def log_level(monkeypatch):
    m = MagicMock()
    monkeypatch.setattr(iceprod.server.server, 'set_log_level', m)
    yield m


@pytest.fixture
def config(monkeypatch):
    m = MagicMock()
    m.return_value = {'logging':{'level':'debug'}}
    monkeypatch.setattr(iceprod.server.server, 'IceProdConfig', m)
    yield m


@pytest.fixture
def queue(monkeypatch):
    m = MagicMock()
    monkeypatch.setattr(iceprod.server.server, 'Queue', m)
    yield m


def test_01_init(config, queue):
    s = Server()
    assert config.called
    assert queue.called


async def test_10_run(config, queue):
    queue.return_value.run = AsyncMock()

    s = Server()
    await s.run()
    assert queue.return_value.run.called == True


def test_90_roll_files(tmp_path):
    filename = str(tmp_path / 'file')
    fd = open(filename, 'ba+')
    fd.write(b'foo')
    fd = roll_files(fd, filename)
    assert os.path.exists(filename)

    files = glob.glob(f'{filename}.*')
    assert len(files) == 1

    fd.write(b'bar')
    fd.close()

    with open(files[0], 'br') as f:
        assert f.read() == b'foo'




"""
Test script for parser
"""

from __future__ import absolute_import, division, print_function

from tests.util import unittest_reporter, glob_tests

import logging
logger = logging.getLogger('parser_test')

import unittest
import json
try:
    import builtins
except ImportError:
    import __builtin__ as builtins


from iceprod.core import parser, dataclasses

class parser_test(unittest.TestCase):
    def setUp(self):
        super(parser_test,self).setUp()

    def tearDown(self):
        super(parser_test,self).tearDown()

    @unittest_reporter
    def test_000_getType(self):
        ret = parser.getType('0')
        self.assertEqual(ret, 0)

        ret = parser.getType('1.24')
        self.assertEqual(ret, 1.24)

        ret = parser.getType('foo')
        self.assertEqual(ret, 'foo')

        ret = parser.getType('"foo"')
        self.assertEqual(ret, '"foo"')

        ret = parser.getType('{"bar":"baz"}')
        self.assertEqual(ret, {"bar":"baz"})

        ret = parser.getType('[1,2,3,4]')
        self.assertEqual(ret, [1,2,3,4])

        ret = parser.getType('{"bar":[1,2,3,4]}')
        self.assertEqual(ret, {"bar":[1,2,3,4]})

        ret = parser.getType('f"oo"')
        self.assertEqual(ret, 'f"oo"')

        ret = parser.getType('f"o\"o"')
        self.assertEqual(ret, 'f"o\"o"')

        ret = parser.getType('f"o\'o"')
        self.assertEqual(ret, 'f"o\'o"')

        input = {'IceModelLocation': '$I3_BUILD/ice-models/resources/models/spice_3.2.1', 'HoleIceParameterization': '$I3_BUILD/ice-models/resources/models/angsens/as.flasher_p1_0.30_p2_-1', 'Perturbations': {'IceWavePlusModes': {'apply': False, 'type': 'default'}, 'Scattering': {'type': 'delta', 'delta': {'x0': [1.0]}}, 'Absorption': {'type': 'delta', 'delta': {'x0': [1.0]}}, 'AnisotropyScale': {'type': 'delta', 'delta': {'x0': [1.0]}}, 'DOMEfficiency': {'type': 'delta', 'delta': {'x0': [1.0]}}, 'HoleIceForward_Unified': {'type': 'delta', 'delta': {'x0': [0.101569, -0.049344]}}}}
        ret = parser.getType(json.dumps(input))
        self.assertEqual(ret, input)

    @unittest_reporter
    def test_001_steering(self):
        """Test parser steering"""
        job = dataclasses.Job()
        job['steering'] = dataclasses.Steering()
        job['steering']['parameters'] = {
            'test': 1,
            'test2': 't2',
            'test3': False,
        }

        p = parser.ExpParser()
        p.job = job

        # run tests
        ret = p.steering_func('test')
        expected = str(job['steering']['parameters']['test'])
        self.assertEqual(ret,expected)

        ret = p.steering_func('test2')
        expected = str(job['steering']['parameters']['test2'])
        self.assertEqual(ret,expected)

        ret = p.steering_func('test3')
        expected = str(job['steering']['parameters']['test3'])
        self.assertEqual(ret,expected)

        with self.assertRaises(parser.GrammarException):
            p.steering_func('test4')

    @unittest_reporter
    def test_002_system(self):
        """Test parser system"""
        job = dataclasses.Job()
        job['steering'] = dataclasses.Steering()
        job['steering']['system'] = {
            'test': 1,
            'test2': 't2',
        }

        p = parser.ExpParser()
        p.job = job

        # run tests
        ret = p.system_func('test')
        expected = str(job['steering']['system']['test'])
        self.assertEqual(ret,expected)

        ret = p.system_func('test2')
        expected = str(job['steering']['system']['test2'])
        self.assertEqual(ret,expected)

        with self.assertRaises(parser.GrammarException):
            p.system_func('test3')

    @unittest_reporter
    def test_003_options(self):
        """Test parser options"""
        job = dataclasses.Job()
        job['options'] = {
            'test': 1,
            'test2': 't2',
        }

        p = parser.ExpParser()
        p.job = job

        # run tests
        ret = p.options_func('test')
        expected = str(job['options']['test'])
        self.assertEqual(ret,expected)

        ret = p.options_func('test2')
        expected = str(job['options']['test2'])
        self.assertEqual(ret,expected)

        with self.assertRaises(parser.GrammarException):
            p.options_func('test3')

    @unittest_reporter
    def test_004_difplus(self):
        """Test parser difplus"""
        job = dataclasses.Job()
        job['difplus'] = dataclasses.DifPlus()
        job['difplus']['dif'] = dataclasses.Dif()
        job['difplus']['plus'] = dataclasses.Plus()
        job['difplus']['plus']['category'] = 'filtered'

        p = parser.ExpParser()
        p.job = job

        # run tests
        ret = p.difplus_func('sensor_name')
        expected = job['difplus']['dif']['sensor_name']
        self.assertEqual(ret,expected)

        ret = p.difplus_func('category')
        expected = job['difplus']['plus']['category']
        self.assertEqual(ret,expected)

        with self.assertRaises(parser.GrammarException):
            p.difplus_func('test')

    @unittest_reporter
    def test_005_eval(self):
        """Test parser eval"""
        p = parser.ExpParser()

        # run tests
        ret = p.eval_func('4+4')
        self.assertEqual(ret, '8')

        ret = p.eval_func('(4+3*2)%3')
        self.assertEqual(ret, '1')

        ret = p.eval_func('2**2')
        self.assertEqual(ret, '4')

        ret = p.eval_func('-1')
        self.assertEqual(ret, '-1')

        ret = p.eval_func('-1')
        self.assertEqual(ret, '-1')

        with self.assertRaises(parser.GrammarException):
            p.eval_func('import os')

        with self.assertRaises(parser.GrammarException):
            p.eval_func('os.remove("/")')

    @unittest_reporter
    def test_006_sprintf(self):
        """Test parser sprintf"""
        p = parser.ExpParser()

        # run tests
        ret = p.sprintf_func('"%d",5')
        self.assertEqual(ret, '5')

        ret = p.sprintf_func('\'%d\',5')
        self.assertEqual(ret, '5')

        ret = p.sprintf_func('%d,5')
        self.assertEqual(ret, '5')

        ret = p.sprintf_func('"%s %06d","testing",12')
        expected = 'testing 000012'
        self.assertEqual(ret,expected)

        ret = p.sprintf_func('"%07d", 12.0000')
        expected = '0000012'
        self.assertEqual(ret,expected)

        ret = p.sprintf_func('"%x", 12')
        self.assertEqual(ret, 'c')

        ret = p.sprintf_func('"%o", 12')
        self.assertEqual(ret, '14')

        with self.assertRaises(parser.GrammarException):
            p.sprintf_func('"%s,12')

        with self.assertRaises(parser.GrammarException):
            p.sprintf_func('%s')

        with self.assertRaises(parser.GrammarException):
            p.sprintf_func('"%s"')

        with self.assertRaises(parser.GrammarException):
            p.sprintf_func('"%f","test"')

    @unittest_reporter
    def test_007_choice(self):
        """Test parser choice"""
        p = parser.ExpParser()

        # run tests
        ret = p.choice_func('1,2,3,4')
        expected = ('1','2','3','4')
        self.assertIn(ret,expected)

        ret = p.choice_func([1,2,3,4])
        self.assertIn(ret,expected)

        ret = p.choice_func('1')
        self.assertEqual(ret, '1')

        with self.assertRaises(parser.GrammarException):
            p.choice_func('')

        with self.assertRaises(parser.GrammarException):
            p.choice_func([])

        with self.assertRaises(parser.GrammarException):
            p.choice_func(123)

    @unittest_reporter
    def test_008_system(self):
        """Test parser environ"""
        job = dataclasses.Job()
        environ = {
            'environment': {
                'test': "blah",
                'test2': 3.14,
                'test3': True,
            },
        }

        p = parser.ExpParser()
        p.job = job
        p.env = environ

        # run tests
        ret = p.environ_func('test')
        self.assertEqual(ret, environ['environment']['test'])

        ret = p.environ_func('test2')
        self.assertEqual(ret, str(environ['environment']['test2']))

        ret = p.environ_func('test3')
        self.assertEqual(ret, str(environ['environment']['test3']))

        with self.assertRaises(parser.GrammarException):
            p.environ_func('test4')

    @unittest_reporter(name='parse() steering')
    def test_010_steering(self):
        """Test parser parse steering"""
        job = dataclasses.Job()
        job['steering'] = dataclasses.Steering()
        job['steering']['parameters'] = {
            'test': 1,
            'test2': 't2',
            'test3': False,
            'test4': '"quotes"',
            'list': [1,2,3,4.0],
            'dict': {'foo':'$steering(test)','bar':{'baz':'$steering(list)'}},
            "SNOWSTORM::config": {
                "IceModelLocation": "$I3_BUILD/ice-models/resources/models/spice_3.2.1",
                "HoleIceParameterization": "$I3_BUILD/ice-models/resources/models/angsens/as.flasher_p1_0.30_p2_-1",
                "Perturbations": {
                    "IceWavePlusModes": {
                        "apply": False,
                        "type": "default"
                    },
                    "Scattering": {
                        "type": "delta",
                        "delta": {
                            "x0": [
                                1.0
                            ]
                        }
                    },
                    "Absorption": {
                        "type": "delta",
                        "delta": {
                            "x0": [
                                1.0
                            ]
                        }
                    },
                    "AnisotropyScale": {
                        "type": "delta",
                        "delta": {
                            "x0": [
                                1.0
                            ]
                        }
                    },
                    "DOMEfficiency": {
                        "type": "delta",
                        "delta": {
                            "x0": [
                                1.0
                            ]
                        }
                    },
                    "HoleIceForward_Unified": {
                        "type": "delta",
                        "delta": {
                            "x0": [
                                0.101569,
                                -0.049344
                            ]
                        }
                    }
                }
            },
        }

        p = parser.ExpParser()

        # run tests
        ret = p.parse('$steering(test)',job=job)
        expected = job['steering']['parameters']['test']
        self.assertEqual(ret,expected)

        ret = p.parse('$steering(test2)',job=job)
        expected = job['steering']['parameters']['test2']
        self.assertEqual(ret,expected)

        ret = p.parse('$steering(test3)',job=job)
        expected = job['steering']['parameters']['test3']
        self.assertEqual(ret,expected)

        ret = p.parse('$steering(test4)',job=job)
        expected = job['steering']['parameters']['test4']
        self.assertEqual(ret,expected)

        ret = p.parse('$steering(list)',job=job)
        expected = job['steering']['parameters']['list']
        self.assertEqual(ret,expected)

        ret = p.parse('$steering(list)[$steering(test)]',job=job)
        expected = job['steering']['parameters']['list'][job['steering']['parameters']['test']]
        self.assertEqual(ret,expected)

        ret = p.parse('$steering(list)[10]',job=job)
        expected = str(job['steering']['parameters']['list'])+'[10]'
        self.assertEqual(ret,expected)

        ret = p.parse('$steering(dict)',job=job)
        expected = {'foo':1,'bar':{'baz':[1,2,3,4.0]}}
        self.assertEqual(ret,expected)

        ret = p.parse('$steering(SNOWSTORM::config)',job=job)
        expected = job['steering']['parameters']['SNOWSTORM::config']
        self.assertEqual(ret,expected)

        for reduction in 'sum', 'len', 'min', 'max':
            ret = p.parse('${}($steering(list))'.format(reduction),job=job)
            expected = getattr(builtins, reduction)(job['steering']['parameters']['list'])
            self.assertEqual(ret,expected)

        ret = p.parse('$steering(nothere)',job=job)
        expected = '$steering(nothere)'
        self.assertEqual(ret,expected)

    @unittest_reporter(name='parse() system')
    def test_011_system(self):
        """Test parser parse system"""
        job = dataclasses.Job()
        job['steering'] = dataclasses.Steering()
        job['steering']['system'] = {
            'test': 1,
            'test2': 't2',
        }

        p = parser.ExpParser()

        # run tests
        ret = p.parse('$system(test)',job=job)
        expected = job['steering']['system']['test']
        self.assertEqual(ret,expected)

        ret = p.parse('$system(test2)',job=job)
        expected = job['steering']['system']['test2']
        self.assertEqual(ret,expected)

        ret = p.parse('$system(test3)',job=job)
        expected = '$system(test3)'
        self.assertEqual(ret,expected)

    @unittest_reporter(name='parse() options')
    def test_012_options(self):
        """Test parser parse options"""
        job = dataclasses.Job()
        job['options'] = {
            'test': 1,
            'test2': 't2',
        }

        p = parser.ExpParser()

        # run tests
        ret = p.parse('$options(test)',job=job)
        expected = job['options']['test']
        self.assertEqual(ret,expected)

        ret = p.parse('$options(test2)',job=job)
        expected = job['options']['test2']
        self.assertEqual(ret,expected)

        ret = p.parse('$(test2)',job=job)
        expected = job['options']['test2']
        self.assertEqual(ret,expected)

        ret = p.parse('$(test3)',job=job)
        expected = '$(test3)'
        self.assertEqual(ret,expected)

        ret = p.parse('$args(test)',job=job)
        expected = job['options']['test']
        self.assertEqual(ret,expected)

    @unittest_reporter(name='parse() difplus')
    def test_013_difplus(self):
        """Test parser difplus"""
        job = dataclasses.Job()
        job['difplus'] = dataclasses.DifPlus()
        job['difplus']['dif'] = dataclasses.Dif()
        job['difplus']['plus'] = dataclasses.Plus()
        job['difplus']['plus']['category'] = 'filtered'

        p = parser.ExpParser()

        # run tests
        ret = p.parse('$metadata(sensor_name)',job=job)
        expected = job['difplus']['dif']['sensor_name']
        self.assertEqual(ret,expected)

        ret = p.parse('$metadata(category)',job=job)
        expected = job['difplus']['plus']['category']
        self.assertEqual(ret,expected)

        ret = p.parse('$metadata(test)',job=job)
        expected = '$metadata(test)'
        self.assertEqual(ret,expected)

    @unittest_reporter(name='parse() eval')
    def test_014_eval(self):
        """Test parser parse eval"""
        p = parser.ExpParser()

        # run tests
        ret = p.parse('$eval(4+4)')
        expected = 8
        self.assertEqual(ret,expected)

        ret = p.parse('$eval(\(4+3*2\)%3)')
        expected = 1
        self.assertEqual(ret,expected)

        ret = p.parse('$eval((4+3*2)%3)')
        expected = 1
        self.assertEqual(ret,expected)

        ret = p.parse('$eval(-1.23)')
        expected = -1.23
        self.assertEqual(ret,expected)

        ret = p.parse('$eval(import os)')
        expected = '$eval(import os)'
        self.assertEqual(ret,expected)

        ret = p.parse('$eval(os.remove\("/"\))')
        expected = '$eval(os.remove("/"))'
        self.assertEqual(ret,expected)

        ret = p.parse('$eval(os.remove("/"))')
        expected = '$eval(os.remove("/"))'
        self.assertEqual(ret,expected)

    @unittest_reporter(name='parse() sprintf')
    def test_015_sprintf(self):
        """Test parser parse sprintf"""
        p = parser.ExpParser()

        # run tests
        ret = p.parse('$sprintf("%d",5)')
        self.assertEqual(ret, 5)

        ret = p.parse('$sprintf(\'%d\',5)')
        self.assertEqual(ret, 5)

        ret = p.parse('$sprintf(%d,5)')
        self.assertEqual(ret, 5)

        ret = p.parse('$sprintf("%s %06d","testing",12)')
        self.assertEqual(ret, 'testing 000012')

        ret = p.parse('$sprintf("%s,12)')
        self.assertEqual(ret, '$sprintf("%s,12)')

        ret = p.parse('$sprintf(%s)')
        self.assertEqual(ret, '$sprintf(%s)')

        ret = p.parse('$sprintf("%s")')
        self.assertEqual(ret, '$sprintf("%s")')

        ret = p.parse('$sprintf("%f","test")')
        self.assertEqual(ret, '$sprintf("%f","test")')

    @unittest_reporter(name='parse() choice')
    def test_016_choice(self):
        """Test parser parse choice"""
        p = parser.ExpParser()

        # run tests
        ret = p.parse('$choice(1,2,3,4)')
        expected = (1,2,3,4)
        self.assertIn(ret,expected)

        ret = p.parse('$choice(1)')
        expected = 1
        self.assertEqual(ret,expected)

        ret = p.parse('$choice()')
        expected = '$choice()'
        self.assertEqual(ret,expected)

    @unittest_reporter(name='parse() environ')
    def test_017_environ(self):
        """Test parser parse environ"""
        p = parser.ExpParser()
        env = {
            'environment':{
                'test': 'blah',
                'test2': 3.14,
                'test3': True,
            }
        }

        # run tests
        ret = p.parse('$environ(test)', env=env)
        self.assertEqual(ret, env['environment']['test'])

        ret = p.parse('$environ(test2)', env=env)
        self.assertEqual(ret, env['environment']['test2'])

        ret = p.parse('$environ(test3)', env=env)
        self.assertEqual(ret, env['environment']['test3'])

    @unittest_reporter(name='parse() env')
    def test_020_env(self):
        """Test parser parse env"""
        p = parser.ExpParser()
        env = {'parameters':{'test':1}}

        # run tests
        ret = p.parse('$(test)',env=env)
        expected = 1
        self.assertEqual(ret,expected)

        ret = p.parse('$test()',env=env)
        expected = '$test()'
        self.assertEqual(ret,expected)

        ret = p.parse('$test',env=env)
        expected = '$test'
        self.assertEqual(ret,expected)

        ret = p.parse('$eval(os.remove("/"))',env=env)
        expected = '$eval(os.remove("/"))'
        self.assertEqual(ret,expected)

    @unittest_reporter(name='parse() job')
    def test_021_job(self):
        """Test parser parse job"""
        job = dataclasses.Job()
        job['test'] = 1
        job['test2'] = 'test'

        p = parser.ExpParser()

        # run tests
        ret = p.parse('$(test)',job=job)
        expected = 1
        self.assertEqual(ret,expected)

        ret = p.parse('$(test2)',job=job)
        expected = 'test'
        self.assertEqual(ret,expected)

        ret = p.parse('ab$(test2)cd',job=job)
        expected = 'abtestcd'
        self.assertEqual(ret,expected)

        ret = p.parse('$(test3)',job=job)
        expected = '$(test3)'
        self.assertEqual(ret,expected)

    @unittest_reporter
    def test_022_parse(self):
        """Test parser parse"""
        job = dataclasses.Job()
        job['steering'] = dataclasses.Steering()
        job['steering']['parameters'] = {
            'FILTER_filtered': 1,
            'SIM_filtered': 2,
        }
        job['steering']['system'] = {
            'test': 'FILTER',
            'test2': 'SIM',
        }
        job['options'] = {
            'FILTER_filtered': 3,
            'SIM_filtered': 4,
        }
        job['difplus'] = dataclasses.DifPlus()
        job['difplus']['dif'] = dataclasses.Dif()
        job['difplus']['plus'] = dataclasses.Plus()
        job['difplus']['plus']['category'] = 'filtered'

        p = parser.ExpParser()

        # run tests
        ret = p.parse('$steering($sprintf("%s_%s",$system(test),$metadata(category)))',job=job)
        expected = 1
        self.assertEqual(ret,expected)

        ret = p.parse('$steering($system(test2)_$metadata(category))',job=job)
        expected = 2
        self.assertEqual(ret,expected)

        ret = p.parse('$options($sprintf("%s_%s",$system(test),$metadata(category)))',job=job)
        expected = 3
        self.assertEqual(ret,expected)

        ret = p.parse('$options($system(test2)_$metadata(category))',job=job)
        expected = 4
        self.assertEqual(ret,expected)

        ret = p.parse('$foo(ba]]r)')
        expected = '$foo(ba]]r)'
        self.assertEqual(ret,expected)

    @unittest_reporter
    def test_030_parse_job_bin(self):
        """Test parsing the job binning. A bug found during prod-test."""
        for j in (0,15,248,1389,10000,10001,20482,83727,493837,1393728):
            job = dataclasses.Job()
            job['steering'] = dataclasses.Steering()
            job['options'] = {
                'job': j,
            }
            p = parser.ExpParser()

            # run tests
            ret = p.parse("$sprintf('%06d-%06d',$eval($(job)//10000*10000),$eval($eval($(job)//10000+1)*10000))",job=job)
            expected = '%06d-%06d'%(j//10000*10000,(j//10000+1)*10000)
            self.assertEqual(ret,expected)

    @unittest_reporter
    def test_031_parse_array_notation(self):
        """Test parsing array notation. A bug found during prod."""
        job = dataclasses.Job()
        job['steering'] = dataclasses.Steering()
        job['steering']['parameters']['array'] = ["foo", "bar", "baz"]
        p = parser.ExpParser()

        # run tests
        for i in range(3):
            ret = p.parse("$steering(array)[{}]".format(i),job=job)
            expected = job['steering']['parameters']['array'][i]
            self.assertEqual(ret,expected)

    @unittest_reporter
    def test_032_parse_dict_notation(self):
        """Test parsing dict notation. A bug found during prod."""
        job = dataclasses.Job()
        job['steering'] = dataclasses.Steering()
        job['steering']['parameters']['dict'] = {"foo":1, "bar":2, "baz":3}
        p = parser.ExpParser()

        # run tests
        for i,n in enumerate(('foo','bar','baz')):
            ret = p.parse("$steering(dict)[{}]".format(n),job=job)
            self.assertEqual(ret,i+1)

    @unittest_reporter
    def test_033_parse_dict_array_notation(self):
        """Test parsing dict+array nested notation. A bug found during prod."""
        job = dataclasses.Job()
        job['steering'] = dataclasses.Steering()
        job['steering']['parameters']['array'] = ["foo", "bar", "baz"]
        job['steering']['parameters']['dict'] = {"foo":1, "bar":2, "baz":3}
        p = parser.ExpParser()

        # run tests
        for i in range(3):
            ret = p.parse("$steering(dict)[$steering(array)[{}]]".format(i),job=job)
            self.assertEqual(ret,i+1)

    @unittest_reporter
    def test_034_parse_dict_array_notation(self):
        """Test parsing dict+array hidden nested notation. A bug found during prod."""
        job = dataclasses.Job()
        job['steering'] = dataclasses.Steering()
        job['steering']['parameters']['array'] = ["foo", "bar", "baz"]
        job['steering']['parameters']['dict'] = {"foo":1, "bar":2, "baz":3}
        job['steering']['parameters']['choose'] = '$steering(array)[0]'
        p = parser.ExpParser()

        ret = p.parse("$steering(dict)[$steering(choose)]",job=job)
        self.assertEqual(ret,1)

    @unittest_reporter
    def test_035_parse_array_with_prefix(self):
        """Test parsing array with a prefix."""
        job = dataclasses.Job()
        job['steering'] = dataclasses.Steering()
        job['steering']['parameters']['array'] = ["foo", "bar", "baz"]
        job['steering']['parameters']['dict'] = {"foo":1, "bar":2, "baz":3}
        job['steering']['parameters']['choose'] = '$steering(array)[0]'
        p = parser.ExpParser()

        ret = p.parse("http://test/$steering(choose)",job=job)
        self.assertEqual(ret,'http://test/foo')

    @unittest_reporter
    def test_036_parse_dict_array_with_prefix(self):
        """Test parsing dict+array hidden nested notation, with a prefix. A bug found during prod."""
        job = dataclasses.Job()
        job['steering'] = dataclasses.Steering()
        job['steering']['parameters']['array'] = ["foo", "bar", "baz"]
        job['steering']['parameters']['dict'] = {"foo":1, "bar":2, "baz":3}
        job['steering']['parameters']['choose'] = '$steering(array)[0]'
        p = parser.ExpParser()

        ret = p.parse("http://test/$steering(dict)[$steering(choose)]",job=job)
        self.assertEqual(ret,'http://test/1')

    @unittest_reporter
    def test_037_parse_sprintf_array(self):
        """Test parsing sprintf+array. A bug found during prod."""
        job = dataclasses.Job()
        job['steering'] = dataclasses.Steering()
        job['steering']['parameters']['array'] = [1, 1.5, 2.2, 3, 5]
        job['steering']['parameters']['choose'] = '$steering(array)[0]'
        p = parser.ExpParser()

        ret = p.parse("$sprintf('%.1fx', $steering(choose))", job=job)
        self.assertEqual(ret, '1.0x')

        ret = p.parse('$sprintf("%.1fx", $(choose))', job=job)
        self.assertEqual(ret, '1.0x')

    @unittest_reporter
    def test_100_scanner(self):
        ret = list(parser.scanner('$foo(bar)'))
        expected = [('starter','$'),('word','foo'),('scopeL','('),('word','bar'),('scopeR',')')]
        self.assertEqual(ret, expected)

        ret = list(parser.scanner('$foo(bar)[baz]'))
        expected = [('starter','$'),('word','foo'),('scopeL','('),('word','bar'),
                    ('scopeR',')'),('bracketL','['),('word','baz'),('bracketR',']')]
        self.assertEqual(ret, expected)

    @unittest_reporter
    def test_110_parser(self):
        ret = list(parser.parser('$foo(bar)'))
        expected = [('starter','$'),('name','foo'),('scopeL','('),('word','bar'),('scopeR',')')]
        self.assertEqual(ret, expected)

        ret = list(parser.parser('$foo(bar)[baz]'))
        expected = [('starter','$'),('name','foo'),('scopeL','('),('word','bar'),
                    ('scopeR',')'),('bracketL','['),('word','baz'),('bracketR',']')]
        self.assertEqual(ret, expected)

        ret = list(parser.parser('$foo(b[a]r)'))
        expected = [('starter','$'),('name','foo'),('scopeL','('),('word','b[a]r'),('scopeR',')')]
        self.assertEqual(ret, expected)

        ret = list(parser.parser('$foo(bar[$(baz)])'))
        expected = [('starter','$'),('name','foo'),('scopeL','('),('word','bar['),
                    ('starter','$'),('name',None),('scopeL','('),
                    ('word','baz'),('scopeR',')'),('word',']'),('scopeR',')')]
        self.assertEqual(ret, expected)

        ret = list(parser.parser('$foo($($(bar)[baz]))'))
        expected = [('starter','$'),('name','foo'),('scopeL','('),('starter','$'),
                    ('name',None),('scopeL','('),('starter','$'),
                    ('name',None),('scopeL','('),('word','bar'),('scopeR',')'),
                    ('bracketL','['),('word','baz'),('bracketR',']'),
                    ('word',''),('scopeR',')'),('word',''),('scopeR',')')]
        self.assertEqual(ret, expected)

        ret = list(parser.parser('$foo($(b[a]r))'))
        expected = [('starter','$'),('name','foo'),('scopeL','('),('starter','$'),
                    ('name',None),('scopeL','('),('word','b[a]r'),('scopeR',')'),
                    ('word',''),('scopeR',')')]
        self.assertEqual(ret, expected)

        ret = list(parser.parser('$foo(bar)[]'))
        expected = [('starter','$'),('name','foo'),('scopeL','('),('word','bar'),('scopeR',')'),
                    ('bracketL','['),('word',''),('bracketR',']'),]
        self.assertEqual(ret, expected)

        ret = list(parser.parser('$foo(ba]]r)'))
        expected = [('starter','$'),('name','foo'),('scopeL','('),('word','ba'),
                    ('bracketR',']'),('word',''),('bracketR',']'),('word','r'),('scopeR',')')]
        self.assertEqual(ret, expected)

        with self.assertRaises(SyntaxError):
            list(parser.parser('$)'))

        ret = list(parser.parser('()'))
        self.assertEqual(ret, [('word','()')])

        with self.assertRaises(SyntaxError):
            list(parser.parser('$(]'))

        with self.assertRaises(SyntaxError):
            list(parser.parser('$foo([]bar)'))

def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    alltests = glob_tests(loader.getTestCaseNames(parser_test))
    suite.addTests(loader.loadTestsFromNames(alltests,parser_test))
    return suite



import asyncio
from collections import defaultdict
from dataclasses import asdict, dataclass
from functools import cached_property
import logging
import os
import re
from typing import Any

from rest_tools.client import RestClient
from tornado.web import HTTPError

from iceprod.core.jsonUtil import json_decode, json_encode
from iceprod.core.parser import ExpParser
from iceprod.common.mongo_queue import Message
from iceprod.core.config import Config as DatasetConfig, ValidationError
from iceprod.services.base import AuthData, BaseAction


logger = logging.getLogger('submit')


TOKEN_PREFIXES = {
    'osdf:///icecube/wipac': 'https://token-issuer.icecube.aq',
    'pelican://osg-htc.org/icecube/wipac': 'https://token-issuer.icecube.aq',
}


SCOPE_RE = re.compile(r'(.*?\$|.*?\d{4,}\-\d{4,}|.*?\/IceCube\/20\d\d\/filtered\/.*?\/[01]\d{3})')


def get_scope(path: str, movement: str) -> str:
    """
    Auto-determines scope based on the path.

    Special cases to trim:
    * subdirectories: example: 000000-000999
    * run months: example: 0123
    * unexpanded variables: anything wth $
    """
    prefix = 'storage.read' if movement == 'input' else 'storage.modify'
    try:
        if match := SCOPE_RE.match(path):
            logger.info('matched scope from RE: %s', match.group(0))
            logger.info('original path: %s', path)
            path = match.group(0)
        path = os.path.dirname(path)
        if not path:
            path = '/'
        path = path.replace('//', '/')
    except Exception:
        logger.warning('error getting scope', exc_info=True)
        raise
    return f'{prefix}:{path}'


type TokenSubmitterTokens = list[Any]


class TokenSubmitter:
    def __init__(self, logger: logging.Logger, cred_client: RestClient, jobs_submitted: int, config: dict[str, Any], username: str, group: str):
        self._logger = logger
        self._cred_client = cred_client
        self._jobs_submitted = jobs_submitted
        self._config = config
        self._username = username
        self._group = group

    @cached_property
    def token_scopes(self) -> dict[str, set[str]]:
        config = self._config.copy()
        parser = ExpParser()

        # get token scopes
        token_scopes = defaultdict(set)
        options = {'jobs_submitted': self._jobs_submitted}
        options.update(config['options'])
        config['options'] = options
        for task in config['tasks']:
            config['options']['task'] = task['name']
            task_token_scopes = defaultdict(set)
            task_data = task['data'].copy()
            for tray in task.get('trays', []):
                task_data.extend(tray.get('data', []).copy())
                for module in tray.get('modules', []):
                    task_data.extend(module.get('data', []).copy())
            for data in task_data:
                self._logger.info('data: %r', data)
                if data['type'] != 'permanent':
                    continue
                remote = parser.parse(data['remote'], job=config)
                for prefix in TOKEN_PREFIXES:
                    if remote.startswith(prefix):
                        if scope := get_scope(remote[len(prefix):], data['movement']):
                            self._logger.info('adding scope %s for remote %s', scope, remote)
                            task_token_scopes[prefix].add(scope)
            # add in manual scopes
            for prefix,scope_str in task['token_scopes'].items():
                for scope in scope_str.split():
                    if scope and '$' not in scope:
                        task_token_scopes[prefix].add(scope)
            # set scopes per task
            for prefix,scopes in task_token_scopes.items():
                sorted_scope_str = ' '.join(sorted(scopes))
                if '$' in sorted_scope_str:
                    raise Exception('bad scope!')
                task['token_scopes'][prefix] = sorted_scope_str
                token_scopes[prefix].update(scopes)

        return token_scopes

    async def get_tokens(self) -> TokenSubmitterTokens:
        token_scopes = self.token_scopes

        if self._group == 'simprod':
            username = 'ice3simusr'
        elif self._group == 'filtering':
            username = 'i3filter'
        else:
            username = self._username

        # request the tokens
        self._logger.info('token requests: %r', dict(token_scopes))
        tokens = []
        for prefix,scopes in token_scopes.items():
            sorted_scope_str = ' '.join(sorted(scopes))
            args = {
                'username': username,
                'scope': sorted_scope_str,
                'url': TOKEN_PREFIXES[prefix],
                'transfer_prefix': prefix,
            }
            try:
                ret = await self._cred_client.request('POST', '/create', args)
            except Exception as e:
                if 'invalid_scope' in str(e):
                    raise Exception(f'Invalid scopes for {prefix}: {sorted_scope_str}')
                raise e
            tokens.append(ret)

        return tokens

    async def submit_tokens(self, dataset_id: str, tokens: TokenSubmitterTokens):
        async with asyncio.TaskGroup() as tg:
            for token in tokens:
                tg.create_task(self._cred_client.request('POST', f'/datasets/{dataset_id}/credentials', token))

    async def tokens_exist(self, dataset_id: str) -> bool:
        try:
            ret = await self._cred_client.request('GET', f'/datasets/{dataset_id}/credentials')
        except Exception:
            self._logger.info('cannot get existing tokens')
            ret = []
        existing_token_scopes = defaultdict(set)
        for tok in ret:
            if tok['transfer_prefix'] in TOKEN_PREFIXES:
                existing_token_scopes[tok['transfer_prefix']].update(tok['scope'].split())

        if self.token_scopes != existing_token_scopes:
            self._logger.warning('existing tokens do not match requested tokens\nexisting: %r\n requested: %r', dict(existing_token_scopes), dict(self.token_scopes))
            return False
        self._logger.info('tokens already exist')
        return True

    async def resubmit_tokens(self, dataset_id: str):
        """
        Resubmit tokens for an existing dataset.
        """
        tokens = await self.get_tokens()
        try:
            await self._cred_client.request('DELETE', f'/datasets/{dataset_id}/credentials')
        except Exception:
            pass
        await self.submit_tokens(dataset_id=dataset_id, tokens=tokens)


@dataclass
class Fields:
    config: str
    description: str
    jobs_submitted: int
    username: str
    group: str = 'users'
    extra_submit_fields: str | None = None
    dataset_id: str = ''


class Action(BaseAction):
    PRIORITY = 10

    async def create(self, args: dict[str, Any], *, auth_data: AuthData) -> str:
        # check auth
        if 'username' not in args:
            args['username'] = auth_data.username
        elif args['username'] != auth_data.username and 'admin' not in auth_data.roles:
            raise HTTPError(400, reason='cannot submit as a different user')

        try:
            data = Fields(**args)
            if isinstance(data.jobs_submitted, str):
                data.jobs_submitted = int(data.jobs_submitted)
        except Exception as e:
            raise HTTPError(400, reason=str(e))
        if data.group not in auth_data.groups:
            raise HTTPError(400, reason='not a member of selected group')

        # validate config
        try:
            dc = DatasetConfig(json_decode(data.config))
            dc.config['version'] = 3.2  # force 3.2 config validation
            dc.fill_defaults()
            dc.validate()
            data.config = json_encode(dc.config)
        except ValidationError as e:
            raise HTTPError(400, reason=str(e))
        except Exception:
            raise HTTPError(400, reason='config is not valid')

        return await self._push(payload=asdict(data), priority=self.PRIORITY)

    async def run(self, message: Message) -> None:
        assert self._api_client and self._cred_client
        data = message.payload

        submit_data = Fields(**data)
        config = json_decode(submit_data.config)

        # generate the tokens and verify permissions
        ts = TokenSubmitter(
            logger=self._logger,
            cred_client=self._cred_client,
            jobs_submitted=submit_data.jobs_submitted,
            config=config,
            username=submit_data.username,
            group=submit_data.group
        )
        tokens = await ts.get_tokens()

        # now submit the dataset
        tasks_per_job = len(config['tasks'])
        ntasks = submit_data.jobs_submitted * tasks_per_job
        args2 = {
            'description': submit_data.description,
            'jobs_submitted': submit_data.jobs_submitted,
            'tasks_submitted': ntasks,
            'tasks_per_job': tasks_per_job,
            'username': submit_data.username,
            'group': submit_data.group,
        }
        if submit_data.extra_submit_fields:
            extra_fields = json_decode(submit_data.extra_submit_fields)
            for name,value in extra_fields.items():
                if name not in args2:
                    args2[name] = value

        ret = await self._api_client.request('POST', '/datasets', args2)
        dataset_id = ret['result']
        await self._api_client.request('PUT', f'/config/{dataset_id}', config)

        # now submit creds
        await ts.submit_tokens(dataset_id=dataset_id, tokens=tokens)

        self._logger.info("submit complete!")

        await self._queue.update_payload(message.uuid, {
            'dataset_id': dataset_id
        })



"""
Test script for i3exec
"""
import argparse
import json
import logging

import iceprod.core.exe
import iceprod.core.i3exec


logger = logging.getLogger('i3exe_test')


async def test_run(tmp_path, mocker):
    config = {
        'tasks': [{
            'name': 'foo',
            'trays': [{
                'modules': [{
                    'env_clear': False,
                    'src': 'foo.py'
                }]
            }],
        }]
    }
    configfile = tmp_path / 'config.json'
    with open(configfile, 'w') as f:
        json.dump(config, f)

    args = argparse.Namespace(
        dataset_id=None,
        task_id=None,
        config=str(configfile),
        task='foo',
        dataset_num=1,
        jobs_submitted=1,
        job_index=0,
        dry_run=0,
    )

    conv = mocker.patch('iceprod.core.exe.WriteToScript.convert', return_value='foo')
    sub = mocker.patch('subprocess.run')
    await iceprod.core.i3exec.run(args)

    conv.assert_called_once()
    sub.assert_called_once_with(['foo'], check=True)

    




from dataclasses import asdict, dataclass
import logging
from typing import Any

from tornado.web import HTTPError

from iceprod.common.mongo_queue import Message
from iceprod.server.states import JOB_STATUS, dataset_prev_statuses
from iceprod.services.actions.submit import TokenSubmitter
from iceprod.services.base import AuthData, BaseAction


logger = logging.getLogger('job_status')


@dataclass
class Fields:
    dataset_id: str
    action: str
    initial_status: str | None = None
    job_ids: list[str] | None = None
    progress: int = 0


class Action(BaseAction):
    PRIORITY = 10

    async def create(self, args: dict[str, Any], *, auth_data: AuthData) -> str:
        try:
            data = Fields(**args)
        except Exception as e:
            raise HTTPError(400, reason=str(e))

        if data.action not in ('hard_reset', 'reset', 'suspend'):
            raise HTTPError(400, reason='invalid action')
        if data.initial_status and data.initial_status not in JOB_STATUS:
            raise HTTPError(400, reason='invalid initial_status')
        if data.initial_status and data.job_ids:
            raise HTTPError(400, reason='cannot define both initial_status and job_ids')

        # check auth
        await self._manual_attr_auth('dataset_id', data.dataset_id, 'write', auth_data=auth_data)

        return await self._push(payload=asdict(data), priority=self.PRIORITY)

    async def run(self, message: Message) -> None:
        assert self._api_client and self._cred_client

        data = Fields(**message.payload)

        if data.action == 'hard_reset':
            task_url = f'/datasets/{data.dataset_id}/task_actions/bulk_hard_reset'
            job_url = f'/datasets/{data.dataset_id}/job_actions/bulk_hard_reset'
        elif data.action == 'reset':
            task_url = f'/datasets/{data.dataset_id}/task_actions/bulk_reset'
            job_url = f'/datasets/{data.dataset_id}/job_actions/bulk_reset'
        elif data.action == 'suspend':
            task_url = f'/datasets/{data.dataset_id}/task_actions/bulk_suspend'
            job_url = f'/datasets/{data.dataset_id}/job_actions/bulk_suspend'
        else:
            raise Exception('invalid set_status')

        dataset = await self._api_client.request('GET', f'/datasets/{data.dataset_id}')

        if data.action in ('reset', 'hard_reset') and dataset['status'] in dataset_prev_statuses('processing'):
            await self._api_client.request('PUT', f'/datasets/{data.dataset_id}/status', {'status': 'processing'})
            config = await self._api_client.request('GET', f'/config/{data.dataset_id}')
            ts = TokenSubmitter(
                logger=self._logger,
                cred_client=self._cred_client,
                jobs_submitted=dataset['jobs_submitted'],
                config=config,
                username=dataset['username'],
                group=dataset['group'],
            )
            if not await ts.tokens_exist(dataset_id=data.dataset_id):
                await ts.resubmit_tokens(dataset_id=data.dataset_id)

        if data.job_ids is None:
            # look up job ids
            job_set = set()
            args = {'keys': 'job_id'}
            if data.initial_status:
                args['status'] = data.initial_status
            ret = await self._api_client.request('GET', f'/datasets/{data.dataset_id}/jobs', args)
            for row in ret.values():
                job_set.add(row['job_id'])
            data.job_ids = list(job_set)

        # specific jobs
        job_ids = data.job_ids.copy()
        while job_ids:
            cur_jobs = job_ids[:5000]
            job_ids = job_ids[5000:]
            await self._api_client.request('POST', task_url, {'jobs': cur_jobs})
            await self._api_client.request('POST', job_url, {'jobs': cur_jobs})
            if job_ids:
                await self._queue.update_payload(message.uuid, {
                    'progress': 100 - len(job_ids)*100//len(data.job_ids)
                })

        self._logger.info("job status update complete!")
        await self._queue.update_payload(message.uuid, {
            'progress': 100
        })



import asyncio
from dataclasses import asdict
from datetime import datetime, timedelta, timezone
import os
import uuid

from iceprod.common.mongo_queue import AsyncMongoQueue, Message

async def test_message():
    now = datetime.fromisoformat('2025-12-12T12:14:01Z')
    m = Message(
        payload={'1': 2},
        uuid=str(uuid.uuid4()),
        status='queued',
        priority=0,
        created_at=now,
    )
    data = {
        'payload': {'1': 2},
        'uuid': m.uuid,
        'status': 'queued',
        'priority': 0,
        'attempts': 0,
        'created_at': now,
    }

    m_dict = asdict(m)
    print(f'{m_dict}')
    for k in data:
        assert m_dict[k] == data[k]


async def test_queue_basics(mongo_url, mongo_clear):
    url = os.environ['DB_URL']
    queue = AsyncMongoQueue(url=url, collection_name='bar')
    await queue.setup()

    # check queue is empty
    ret = await queue.pop()
    assert not ret

    ret = await queue.count()
    assert ret == 0

    # put data on the queue
    data = {'1': 2, '3': 4}
    id_ = await queue.push(data)

    ret = await queue.count()
    assert ret == 1

    ret2 = await queue.get_status(id_)
    assert ret2 == 'queued'

    # get it from the queue
    ret = await queue.pop()
    assert ret
    assert ret.payload == data

    ret2 = await queue.get_status(id_)
    assert ret2 == 'processing'

    # check that there's only the one message
    ret2 = await queue.pop()
    assert not ret2

    # put it back
    await queue.release(ret.uuid)

    ret2 = await queue.get_status(id_)
    assert ret2 == 'queued'

    # now get it again
    ret = await queue.pop()
    assert ret
    assert ret.payload == data

    # check that it's complete
    await queue.complete(id_)
    ret = await queue.get_status(id_)
    assert ret == 'complete'

    ret = await queue.pop()
    assert ret == None

    # check that it's failed
    await queue.fail(id_, error_message='test')
    ret2 = await queue.get_status(id_)
    assert ret2 == 'error'

    ret = await queue.pop()
    assert ret == None

    ret2 = await queue.get_error(id_)
    assert ret2 == 'test'

    await queue.close()


async def test_queue_dedup(mongo_url, mongo_clear):
    url = os.environ['DB_URL']
    queue = AsyncMongoQueue(url=url, collection_name='bar')
    await queue.setup()

    id_ = await queue.push_if_not_exists({'1': 1, '2': 2})

    id2_ = await queue.push_if_not_exists({'1': 1, '2': 2})
    assert id_ == id2_

    id2_ = await queue.push_if_not_exists({'1': 1, '2': 2}, {'1': 1})
    assert id_ == id2_

    id2_ = await queue.push_if_not_exists({'2': 2})
    assert id_ == id2_

    ret = await queue.pop()
    assert ret
    assert ret.uuid == id_

    ret = await queue.pop()
    assert not ret

    await queue.close()



async def test_queue_retries(mongo_url, mongo_clear):
    url = os.environ['DB_URL']
    queue = AsyncMongoQueue(url=url, collection_name='bar')
    await queue.setup()

    id_ = await queue.push({'1': 1})

    ret = await queue.pop(timeout_seconds=.01)
    assert ret
    assert ret.uuid == id_
    
    ret = await queue.pop(timeout_seconds=.01)
    assert not ret

    await asyncio.sleep(.01)
    
    # check queue retry
    ret = await queue.pop(timeout_seconds=.01)
    assert ret
    assert ret.uuid == id_
    assert ret.attempts == 2

    await queue.close()


async def test_queue_context(mongo_url, mongo_clear):
    url = os.environ['DB_URL']
    queue = AsyncMongoQueue(url=url, collection_name='bar')
    await queue.setup()

    data = {'1': 2, '3': 4}
    id_ = await queue.push(data)

    try:
        async with queue.process_next() as ret:
            assert ret
            assert ret == data

            # check that there's only the one message
            ret2 = await queue.pop()
            assert not ret2
    except RuntimeError:
        pass

    # check that it's not poppable anymore
    ret2 = await queue.pop()
    assert not ret2

    ret2 = await queue.get_status(id_)
    assert ret2 == 'complete'

    await queue.close()


async def test_queue_context_error(mongo_url, mongo_clear):
    url = os.environ['DB_URL']
    queue = AsyncMongoQueue(url=url, collection_name='bar')
    await queue.setup()

    data = {'1': 2, '3': 4}
    id_ = await queue.push(data)

    try:
        async with queue.process_next() as ret:
            assert ret
            assert ret == data

            # check that there's only the one message
            ret2 = await queue.pop()
            assert not ret2

            # put it back
            raise RuntimeError()
    except RuntimeError:
        pass

    # check that it's not poppable anymore
    ret2 = await queue.pop()
    assert not ret2

    ret2 = await queue.get_status(id_)
    assert ret2 == 'error'

    await queue.close()


async def test_queue_extras(mongo_url, mongo_clear):
    url = os.environ['DB_URL']
    queue = AsyncMongoQueue(url=url, collection_name='bar', extra_indexes={
        'dataset_id': {
            'keys': [('dataset_id', 1)]
        }
    })
    await queue.setup()

    data = {'dataset_id': 2, '3': 4}
    id_ = await queue.push(data)

    ret = await queue.lookup_by_payload({'dataset_id': 2})
    assert ret
    assert ret.uuid == id_



async def test_queue_sort(mongo_url, mongo_clear):
    url = os.environ['DB_URL']
    queue = AsyncMongoQueue(url=url, collection_name='bar')
    await queue.setup()

    u1 = await queue.push({'1': 1})
    u2 = await queue.push({'2': 1})
    u3 = await queue.push({'3': 1})
    u4 = await queue.push({'4': 1})

    ret = await queue.pop()
    assert ret
    assert ret.uuid == u1

    ret = await queue.pop()
    assert ret
    assert ret.uuid == u2

    await queue.release(u1)

    ret = await queue.pop()
    assert ret
    assert ret.uuid == u3

    ret = await queue.pop()
    assert ret
    assert ret.uuid == u4
    await queue.release(u4)

    ret = await queue.pop()
    assert ret
    assert ret.uuid == u1
    await queue.release(u1)

    await queue.release(u1)
    await queue.release(u1)
    await queue.release(u2)

    ret = await queue.pop()
    assert ret
    assert ret.uuid == u2

    await queue.close()


from prometheus_client import REGISTRY, GC_COLLECTOR, PLATFORM_COLLECTOR, PROCESS_COLLECTOR
import pytest


# disable these for testing
REGISTRY.unregister(GC_COLLECTOR)
REGISTRY.unregister(PLATFORM_COLLECTOR)
REGISTRY.unregister(PROCESS_COLLECTOR)


@pytest.fixture(autouse=True)
def clear_registry():
    collectors = tuple(REGISTRY._collector_to_names.keys())
    for collector in collectors:
        REGISTRY.unregister(collector)
    yield


@pytest.fixture(autouse=True)
def i3prod_path(monkeypatch, tmp_path):
    (tmp_path / 'etc').mkdir(mode=0o700)
    monkeypatch.setenv('I3PROD', str(tmp_path))
    yield tmp_path



import pytest

from rest_tools.client import RestClient

from iceprod.scheduled_tasks import buffer_jobs_tasks


async def test_buffer_jobs_tasks_run(requests_mock):
    requests_mock.get('http://iceprod.test/datasets', json={'123': {}, '456': {}})
    requests_mock.post('http://iceprod.test/actions/materialization', json={'result': 'matid'})
    requests_mock.get('http://iceprod.test/actions/materialization/matid', json={'status': 'complete'})
    
    rc = RestClient('http://iceprod.test')
    await buffer_jobs_tasks.run(rc, delay=0)


async def test_buffer_jobs_tasks_run_dataset(requests_mock):
    requests_mock.post('http://iceprod.test/actions/materialization', json={'result': 'matid'})
    requests_mock.get('http://iceprod.test/actions/materialization/matid', json={'status': 'complete'})
    
    rc = RestClient('http://iceprod.test')
    await buffer_jobs_tasks.run(rc, only_dataset='12345', delay=0)


async def test_buffer_jobs_tasks_run_fail(requests_mock):
    requests_mock.get('http://iceprod.test/datasets', json={'123': {}, '456': {}})
    requests_mock.post('http://iceprod.test/actions/materialization', json={'result': 'matid'})
    requests_mock.get('http://iceprod.test/actions/materialization/matid', json={'id': 'matid', 'status': 'error'})

    rc = RestClient('http://iceprod.test')
    with pytest.raises(Exception):
        await buffer_jobs_tasks.run(rc, debug=True, delay=0)
